import time
from datetime import date, datetime, timedelta
from typing import Dict, Any, Optional

from dateutil.parser import parse as dateutil_parse
from loguru import logger
from sqlalchemy import Column, Date, DateTime, Float, Integer, Table, Unicode, func
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm import relation, Session
from sqlalchemy.schema import ForeignKey

from flexget import db_schema, plugin
from flexget.event import event
from flexget.utils import requests
from flexget.utils.database import year_property
from flexget.utils.soup import get_soup

logger = logger.bind(name='api_bluray')
Base = db_schema.versioned_base('api_bluray', 0)

# association tables
genres_table = Table(
    'bluray_movie_genres',
    Base.metadata,
    Column('movie_id', Integer, ForeignKey('bluray_movies.id')),
    Column('genre_name', Integer, ForeignKey('bluray_genres.name')),
)
Base.register_table(genres_table)

BASE_URL = 'http://m.blu-ray.com/'


def bluray_request(endpoint, **params) -> Any:
    full_url = BASE_URL + endpoint
    response = requests.get(full_url, params=params)
    if response.content:
        return response.json(strict=False)


def extract_release_date(bluray_entry: Dict[str, Any]) -> date:
    release_date = bluray_entry.get('reldate')
    if not release_date or release_date.lower() == 'no release date':
        try:
            year = int(bluray_entry.get('year'))
        except (TypeError, ValueError):
            pass
        else:
            return date(year, 12, 31)
        return datetime.now().date().replace(month=12, day=31)
    return dateutil_parse(release_date).date()


class BlurayMovie(Base):
    __tablename__ = 'bluray_movies'

    id = Column(Integer, primary_key=True, autoincrement=False, nullable=False)
    name = Column(Unicode)
    url = Column(Unicode)
    release_date = Column(Date)
    year = year_property('release_date')
    runtime = Column(Integer)
    overview = Column(Unicode)
    country = Column(Unicode)
    studio = Column(Unicode)
    rating = Column(Float)
    bluray_rating = Column(Integer)
    certification = Column(Unicode)
    _genres = relation('BlurayGenre', secondary=genres_table, backref='movies')
    genres = association_proxy('_genres', 'name')
    updated = Column(DateTime, default=datetime.now, nullable=False)

    def __init__(self, title: str, year: Optional[int]) -> None:
        if year:
            title_year = f'{title} ({year})'
        else:
            title_year = title

        params = {
            'section': 'bluraymovies',
            'country': 'ALL',
            'keyword': title,
            '_': str(int(time.time() * 1000)),
        }

        country_params = {'_': params['_']}
        try:
            response = bluray_request('quicksearch/search.php', **params)

            if not response or 'items' not in response:
                raise LookupError(
                    'No search results found for {} on blu-ray.com'.format(title_year)
                )

            search_results = response['items']
            countries = bluray_request('countries.json.php', **country_params) or {}

            search_results = sorted(search_results, key=lambda k: extract_release_date(k))
        except requests.RequestException as e:
            raise LookupError('Error searching for {} on blu-ray.com: {}'.format(title_year, e))

        # Simply take the first result unless year does not match
        for result in search_results:
            if year and str(year) != result['year']:
                continue

            self.id = int(result['url'].split('/')[-2])
            self.name = result['title']

            flag = result['flag']
            country_code = (
                flag.split('/')[-1].split('.')[0].lower()
            )  # eg. http://some/url/UK.png -> uk
            # find country based on flag url, default United States
            country = 'United States'
            for c in countries['countries']:
                if c['c'].lower() == country_code:
                    country = c['n']
            self.country = country
            self.release_date = extract_release_date(result)
            self.bluray_rating = int(result['rating']) if result['rating'] else None

            # Used for parsing some more data, sadly with soup
            self.url = result['url']

            try:
                movie_info_response = requests.get(self.url).content
            except (requests.RequestException, ConnectionError) as e:
                raise LookupError("Couldn't connect to blu-ray.com. %s" % self.url)

            movie_info = get_soup(movie_info_response)

            # runtime and rating, should be the last span tag with class subheading
            bluray_info = movie_info.find('div', attrs={'class': 'bluray'})
            bluray_info = bluray_info.find_all('span', attrs={'class': 'subheading'})[
                -1
            ].text.split('|')

            self.studio = bluray_info[0].strip()

            for info in bluray_info[1:]:
                if 'min' in info:
                    self.runtime = int(info.replace('min', '').strip())
                elif 'Rated' in info:
                    self.certification = info.replace('Rated', '').strip()

            # rating
            rating_tag = movie_info.find('div', id='ratingscore')
            self.rating = float(rating_tag.text.strip()) if rating_tag else None

            # Third onecol_content contains some information we want
            onecol_content = movie_info.find_all('div', attrs={'class': 'onecol_content'})[2]

            # overview, genres etc
            contents = onecol_content.find('div').find('div')

            overview_tag = contents.find('p')
            self.overview = overview_tag.text.strip() if overview_tag else None

            # genres
            genres_table = contents.find('table')
            if not genres_table:
                break

            genres_content = genres_table.find_all('tr')
            if not genres_content:
                break

            genres = set()
            for genre in genres_content:
                genres.add(genre.find('td').text.strip())
            self._genres = [BlurayGenre(name=genre) for genre in genres]
            break
        else:
            raise LookupError('No search results found for {} on blu-ray.com'.format(title_year))


class BlurayGenre(Base):
    __tablename__ = 'bluray_genres'

    name = Column(Unicode, primary_key=True, nullable=False)


class BluraySearchResult(Base):
    __tablename__ = 'bluray_search_results'

    search = Column(Unicode, primary_key=True)
    movie_id = Column(Integer, ForeignKey('bluray_movies.id'), nullable=True)
    movie = relation(BlurayMovie)

    def __init__(
        self, search: str, movie_id: Optional[int] = None, movie: Optional[BlurayMovie] = None
    ) -> None:
        self.search = search.lower()
        if movie_id:
            self.movie_id = movie_id
        if movie:
            self.movie = movie


class ApiBluray:
    """Does lookups to Blu-ray.com and provides movie information. Caches lookups."""

    @staticmethod
    def lookup(
        title: Optional[str] = None,
        year: Optional[int] = None,
        only_cached: bool = False,
        session: Optional[Session] = None,
    ):
        if not title:
            raise LookupError('No criteria specified for blu-ray.com lookup')
        title_year = title + ' ({})'.format(year) if year else title

        movie_filter = session.query(BlurayMovie).filter(
            func.lower(BlurayMovie.name) == title.lower()
        )
        if year:
            movie_filter = movie_filter.filter(BlurayMovie.year == year)
        movie = movie_filter.first()
        if not movie:
            found = (
                session.query(BluraySearchResult)
                .filter(BluraySearchResult.search == title_year.lower())
                .first()
            )
            if found and found.movie:
                movie = found.movie

        if movie:
            # Movie found in cache, check if cache has expired. Shamefully stolen from api_tmdb
            refresh_time = timedelta(days=2)
            if movie.release_date:
                if movie.release_date > datetime.now().date() - timedelta(days=7):
                    # Movie is less than a week old, expire after 1 day
                    refresh_time = timedelta(days=1)
                else:
                    age_in_years = (datetime.now().date() - movie.release_date).days / 365
                    refresh_time += timedelta(days=age_in_years * 5)
            if movie.updated < datetime.now() - refresh_time and not only_cached:
                logger.debug(
                    'Cache has expired for {}, attempting to refresh from blu-ray.com.', movie.name
                )
                try:
                    updated_movie = BlurayMovie(title=title, year=year)
                except LookupError as e:
                    logger.error(
                        'Error refreshing movie details for {} from blu-ray.com, cached info being used. {}',
                        title,
                        e,
                    )
                else:
                    movie = session.merge(updated_movie)
            else:
                logger.debug('Movie {} information restored from cache.', movie.name)
        else:
            if only_cached:
                raise LookupError('Movie %s not found from cache' % title_year)
            # There was no movie found in the cache, do a lookup from blu-ray.com
            logger.verbose('Searching from blu-ray.com `{}`', title)

            # Add/merge movie to db
            movie = BlurayMovie(title=title, year=year)

            # Add to search results table if necessary
            if title.lower() != movie.name.lower():
                session.add(BluraySearchResult(search=title_year.lower(), movie_id=movie.id))

            session.merge(movie)

            if not movie:
                raise LookupError('Unable to find movie on blu-ray: {}'.format(title_year))

        return movie


@event('plugin.register')
def register_plugin() -> None:
    plugin.register(ApiBluray, 'api_bluray', api_ver=2, interfaces=[])



from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='generate')


class Generate:
    """Generates n number of random entries. Used for debugging purposes."""

    schema = {'type': 'integer'}

    def on_task_input(self, task, config):
        amount = config or 0  # hackily makes sure it's an int value
        entries = []
        for i in range(amount):
            entry = Entry()
            import random
            import string

            entry['url'] = 'http://localhost/generate/%s/%s' % (
                i,
                ''.join(
                    [random.choice(string.ascii_letters + string.digits) for x in range(1, 30)]
                ),
            )
            entry['title'] = ''.join(
                [random.choice(string.ascii_letters + string.digits) for x in range(1, 30)]
            )
            entry['description'] = ''.join(
                [random.choice(string.ascii_letters + string.digits) for x in range(1, 1000)]
            )
            entries.append(entry)
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(Generate, 'generate', api_ver=2, debug=True)



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.log import log_once

logger = logger.bind(name='imdb')


class FilterImdb:
    """
    This plugin allows filtering based on IMDB score, votes and genres etc.

    Note: All parameters are optional. Some are mutually exclusive.

    Configuration::

      min_score: <num>
      min_votes: <num>
      min_meta_score: <num>
      min_year: <num>
      max_year: <num>

      # accept movies with any of these genres
      accept_genres:
        - genre1
        - genre2

      # reject if genre contains any of these
      reject_genres:
        - genre1
        - genre2

      # reject if language contain any of these
      reject_languages:
        - language1

      # accept only these primary languages
      accept_languages:
        - language1

      # accept movies with any of these actors
      accept_actors:
        - nm0004695
        - nm0004754

      # reject movie if it has any of these actors
      reject_actors:
        - nm0001191
        - nm0002071

      # accept all movies by these directors
      accept_directors:
        - nm0000318

      # reject movies by these directors
      reject_directors:
        - nm0093051

      # accept all movies by these writers
      accept_writers:
        - nm0000318

      # reject movies by these writers
      reject_writers:
        - nm0093051

      # reject movies/TV shows with any of these ratings
      reject_mpaa_ratings:
        - PG_13
        - R
        - X

      # accept movies/TV shows with only these ratings
      accept_mpaa_ratings:
        - PG
        - G
        - TV_Y
    """

    schema = {
        'type': 'object',
        'properties': {
            'min_year': {'type': 'integer'},
            'max_year': {'type': 'integer'},
            'min_votes': {'type': 'integer'},
            'min_meta_score': {'type': 'integer'},
            'min_score': {'type': 'number'},
            'accept_genres': {'type': 'array', 'items': {'type': 'string'}},
            'reject_genres': {'type': 'array', 'items': {'type': 'string'}},
            'reject_languages': {'type': 'array', 'items': {'type': 'string'}},
            'accept_languages': {'type': 'array', 'items': {'type': 'string'}},
            'reject_actors': {'type': 'array', 'items': {'type': 'string'}},
            'accept_actors': {'type': 'array', 'items': {'type': 'string'}},
            'reject_directors': {'type': 'array', 'items': {'type': 'string'}},
            'accept_directors': {'type': 'array', 'items': {'type': 'string'}},
            'reject_writers': {'type': 'array', 'items': {'type': 'string'}},
            'accept_writers': {'type': 'array', 'items': {'type': 'string'}},
            'reject_mpaa_ratings': {'type': 'array', 'items': {'type': 'string'}},
            'accept_mpaa_ratings': {'type': 'array', 'items': {'type': 'string'}},
        },
        'additionalProperties': False,
    }

    # Run later to avoid unnecessary lookups
    @plugin.priority(120)
    def on_task_filter(self, task, config):

        lookup = plugin.get('imdb_lookup', self).lookup

        # since the plugin does not reject anything, no sense going trough accepted
        for entry in task.undecided:

            force_accept = False

            try:
                lookup(entry)
            except plugin.PluginError as e:
                # logs skip message once trough log_once (info) and then only when ran from cmd line (w/o --cron)
                log_once(
                    'Skipping %s because of an error: %s' % (entry['title'], e.value),
                    logger=logger,
                )
                continue

            # for key, value in entry.iteritems():
            #     log.debug('%s = %s (type: %s)' % (key, value, type(value)))

            # Check defined conditions, TODO: rewrite into functions?
            reasons = []
            if 'min_score' in config:
                if entry.get('imdb_score', 0) < config['min_score']:
                    reasons.append(
                        'min_score (%s < %s)' % (entry.get('imdb_score'), config['min_score'])
                    )
            if 'min_votes' in config:
                if entry.get('imdb_votes', 0) < config['min_votes']:
                    reasons.append(
                        'min_votes (%s < %s)' % (entry.get('imdb_votes'), config['min_votes'])
                    )
            if 'min_meta_score' in config:
                if entry.get('imdb_meta_score', 0) < config['min_meta_score']:
                    reasons.append(
                        'min_meta_score (%s < %s)'
                        % (entry.get('imdb_meta_score'), config['min_meta_score'])
                    )
            if 'min_year' in config:
                if entry.get('imdb_year', 0) < config['min_year']:
                    reasons.append(
                        'min_year (%s < %s)' % (entry.get('imdb_year'), config['min_year'])
                    )
            if 'max_year' in config:
                if entry.get('imdb_year', 0) > config['max_year']:
                    reasons.append(
                        'max_year (%s > %s)' % (entry.get('imdb_year'), config['max_year'])
                    )

            if 'accept_genres' in config:
                accepted = config['accept_genres']
                accept_genre = False
                for genre in entry.get('imdb_genres', []):
                    if genre in accepted:
                        accept_genre = True
                        break
                if accept_genre == False:
                    reasons.append('accept_genres')

            if 'reject_genres' in config:
                rejected = config['reject_genres']
                for genre in entry.get('imdb_genres', []):
                    if genre in rejected:
                        reasons.append('reject_genres')
                        break

            if 'reject_languages' in config:
                rejected = config['reject_languages']
                for language in entry.get('imdb_languages', []):
                    if language in rejected:
                        reasons.append('reject_languages')
                        break

            if 'accept_languages' in config:
                accepted = config['accept_languages']
                if entry.get('imdb_languages') and entry['imdb_languages'][0] not in accepted:
                    # Reject if the first (primary) language is not among acceptable languages
                    reasons.append('accept_languages')

            if 'reject_actors' in config:
                rejected = config['reject_actors']
                for actor_id, actor_name in entry.get('imdb_actors', {}).items():
                    if actor_id in rejected or actor_name in rejected:
                        reasons.append('reject_actors %s' % actor_name or actor_id)
                        break

            # Accept if actors contains an accepted actor, but don't reject otherwise
            if 'accept_actors' in config:
                accepted = config['accept_actors']
                for actor_id, actor_name in entry.get('imdb_actors', {}).items():
                    if actor_id in accepted or actor_name in accepted:
                        logger.debug(
                            'Accepting because of accept_actors {}', actor_name or actor_id
                        )
                        force_accept = True
                        break

            if 'reject_directors' in config:
                rejected = config['reject_directors']
                for director_id, director_name in entry.get('imdb_directors', {}).items():
                    if director_id in rejected or director_name in rejected:
                        reasons.append('reject_directors %s' % director_name or director_id)
                        break

            # Accept if the director is in the accept list, but do not reject if the director is unknown
            if 'accept_directors' in config:
                accepted = config['accept_directors']
                for director_id, director_name in entry.get('imdb_directors', {}).items():
                    if director_id in accepted or director_name in accepted:
                        logger.debug(
                            'Accepting because of accept_directors {}',
                            director_name or director_id,
                        )
                        force_accept = True
                        break

            if 'reject_writers' in config:
                rejected = config['reject_writers']
                for writer_id, writer_name in entry.get('imdb_writers', {}).items():
                    if writer_id in rejected or writer_name in rejected:
                        reasons.append('reject_writers %s' % writer_name or writer_id)
                        break

            # Accept if the writer is in the accept list, but do not reject if the writer is unknown
            if 'accept_writers' in config:
                accepted = config['accept_writers']
                for writer_id, writer_name in entry.get('imdb_writers', {}).items():
                    if writer_id in accepted or writer_name in accepted:
                        logger.debug(
                            'Accepting because of accept_writers {}', writer_name or writer_id
                        )
                        force_accept = True
                        break

            if 'reject_mpaa_ratings' in config:
                rejected = config['reject_mpaa_ratings']
                if entry.get('imdb_mpaa_rating') in rejected:
                    reasons.append('reject_mpaa_ratings %s' % entry['imdb_mpaa_rating'])

            if 'accept_mpaa_ratings' in config:
                accepted = config['accept_mpaa_ratings']
                if entry.get('imdb_mpaa_rating') not in accepted:
                    reasons.append('accept_mpaa_ratings %s' % entry.get('imdb_mpaa_rating'))

            if reasons and not force_accept:
                msg = 'Didn\'t accept `%s` because of rule(s) %s' % (
                    entry.get('imdb_name', None) or entry['title'],
                    ', '.join(reasons),
                )
                if task.options.debug:
                    logger.debug(msg)
                else:
                    if task.options.cron:
                        log_once(msg, logger)
                    else:
                        logger.info(msg)
            else:
                logger.debug('Accepting {}', entry['title'])
                entry.accept()


@event('plugin.register')
def register_plugin():
    plugin.register(FilterImdb, 'imdb', api_ver=2)



import os

from flask import Flask, render_template, send_from_directory
from flask_compress import Compress
from loguru import logger

from flexget.webserver import register_app

logger = logger.bind(name='webui')

config = None
exists = True

ui_base = os.path.dirname(os.path.realpath(__file__))
ui_dist = os.path.join(ui_base, 'dist')
ui_assets = os.path.join(ui_dist, 'assets')

webui_app = Flask(__name__, template_folder=ui_dist)
Compress(webui_app)
webui_app.url_path = '/'


@webui_app.route('/assets/<path:path>')
def serve_app(path):
    return send_from_directory(ui_assets, path)


@webui_app.route('/')
@webui_app.route('/<path:path>')
def root(path='index.html'):
    if not exists:
        return send_from_directory(ui_base, 'index.html')
    return render_template('index.html', base_url=config['base_url'])


def register_web_ui(cfg):
    global config, exists
    config = cfg

    if not os.path.exists(ui_dist):
        exists = False
        logger.critical(
            'Failed to start web ui,'
            ' this can happen if you are running from GitHub version and forgot to run the web ui build, '
            'see https://flexget.com/Web-UI for instructions'
        )

    register_app(webui_app.url_path, webui_app, 'WebUI (v2)')



import copy

from flask import jsonify, redirect
from flask_restx import inputs

from flexget.api import APIResource, api
from flexget.api.app import NotFoundError, etag

from . import db
from .api_trakt import ApiTrakt

trakt_api = api.namespace('trakt', description='Trakt lookup endpoint')


class ObjectsContainer:
    translation_object = {
        'type': 'object',
        'patternProperties': {
            '^[/d]$': {
                'type': 'object',
                'properties': {
                    'overview': {'type': 'string'},
                    'tagline': {'type': 'string'},
                    'title': {'type': 'string'},
                },
                'required': ['overview', 'tagline', 'title'],
                'additionalProperties': False,
            }
        },
    }

    actor_object = {
        'type': 'object',
        'patternProperties': {
            '^[/d]$': {
                'type': 'object',
                'properties': {
                    'imdb_id': {'type': 'string'},
                    'name': {'type': 'string'},
                    'tmdb_id': {'type': 'integer'},
                    'trakt_id': {'type': 'integer'},
                    'trakt_slug': {'type': 'string'},
                    'birthday': {'type': 'string'},
                    'biography': {'type': ['string', 'null']},
                    'homepage': {'type': 'string'},
                    'death': {'type': ['string', 'null']},
                },
                'required': [
                    'imdb_id',
                    'name',
                    'tmdb_id',
                    'trakt_id',
                    'trakt_slug',
                    'birthday',
                    'biography',
                    'homepage',
                    'death',
                ],
                'additionalProperties': False,
            }
        },
    }

    base_return_object = {
        'type': 'object',
        'properties': {
            'translations': translation_object,
            'actors': actor_object,
            'cached_at': {'type': 'string', 'format': 'date-time'},
            'genres': {'type': 'array', 'items': {'type': 'string'}},
            'id': {'type': 'integer'},
            'overview': {'type': ['string', 'null']},
            'runtime': {'type': ['integer', 'null']},
            'rating': {'type': ['number', 'null']},
            'votes': {'type': ['integer', 'null']},
            'language': {'type': ['string', 'null']},
            'updated_at': {'type': 'string', 'format': 'date-time'},
            'title': {'type': 'string'},
            'year': {'type': ['integer', 'null']},
            'homepage': {'type': ['string', 'null']},
            'slug': {'type': ['string', 'null']},
            'tmdb_id': {'type': ['integer', 'null']},
            'imdb_id': {'type': ['string', 'null']},
        },
        'required': [
            'cached_at',
            'genres',
            'id',
            'overview',
            'runtime',
            'rating',
            'votes',
            'language',
            'updated_at',
            'title',
            'year',
            'homepage',
            'slug',
            'tmdb_id',
            'imdb_id',
        ],
        'additionalProperties': False,
    }

    series_return_object = copy.deepcopy(base_return_object)
    series_return_object['properties']['tvdb_id'] = {'type': ['integer', 'null']}
    series_return_object['properties']['tvrage_id'] = {'type': ['integer', 'null']}
    series_return_object['properties']['first_aired'] = {
        'type': ['string', 'null'],
        'format': 'date-time',
    }
    series_return_object['properties']['air_day'] = {'type': ['string', 'null']}
    series_return_object['properties']['air_time'] = {'type': ['string', 'null']}
    series_return_object['properties']['certification'] = {'type': ['string', 'null']}
    series_return_object['properties']['network'] = {'type': ['string', 'null']}
    series_return_object['properties']['country'] = {'type': ['string', 'null']}
    series_return_object['properties']['status'] = {'type': 'string'}
    series_return_object['properties']['timezone'] = {'type': ['string', 'null']}
    series_return_object['properties']['number_of_aired_episodes'] = {'type': ['integer', 'null']}
    series_return_object['required'] += [
        'tvdb_id',
        'tvrage_id',
        'first_aired',
        'air_day',
        'air_time',
        'certification',
        'network',
        'country',
        'status',
        'timezone',
        'number_of_aired_episodes',
    ]

    movie_return_object = copy.deepcopy(base_return_object)
    movie_return_object['properties']['tagline'] = {'type': 'string'}
    movie_return_object['properties']['released'] = {'type': 'string'}
    movie_return_object['properties']['trailer'] = {'type': ['string', 'null']}
    movie_return_object['required'] += ['tagline', 'released', 'trailer']


series_return_schema = api.schema_model(
    'series_return_schema', ObjectsContainer.series_return_object
)
movie_return_schema = api.schema_model('movie_return_schema', ObjectsContainer.movie_return_object)

lookup_parser = api.parser()
lookup_parser.add_argument('title', help='Movie Name')
lookup_parser.add_argument('year', type=int, help='Lookup year')
lookup_parser.add_argument('trakt_id', type=int, help='Trakt ID')
lookup_parser.add_argument('trakt_slug', help='Trakt slug')
lookup_parser.add_argument('tmdb_id', type=int, help='TMDB ID')
lookup_parser.add_argument('imdb_id', help='IMDB ID')
lookup_parser.add_argument('tvdb_id', type=int, help='TVDB ID')
lookup_parser.add_argument('tvrage_id', type=int, help='TVRage ID')
lookup_parser.add_argument(
    'include_actors', type=inputs.boolean, help='Include actors in response'
)
lookup_parser.add_argument(
    'include_translations', type=inputs.boolean, help='Include translations in response'
)


@trakt_api.route('/series/')
class TraktSeriesSearchApi(APIResource):
    @etag(cache_age=3600)
    @api.response(200, 'Successfully found show', series_return_schema)
    @api.response(NotFoundError)
    @api.doc(parser=lookup_parser)
    def get(self, session=None):
        """Trakt series lookup"""
        args = lookup_parser.parse_args()
        include_actors = args.pop('include_actors')
        include_translations = args.pop('include_translations')
        kwargs = args
        try:
            series = ApiTrakt.lookup_series(session=session, **kwargs)
        except LookupError as e:
            raise NotFoundError(e.args[0])
        result = series.to_dict()
        if include_actors:
            result['actors'] = db.list_actors(series.actors)
        if include_translations:
            result['translations'] = db.get_translations_dict(series.translations, 'show')
        return jsonify(result)


@trakt_api.route('/series/<string:title>')
@api.doc(params={'title': 'Series name'})
class TraktSeriesWithTitleSearchApi(APIResource):
    @api.doc(parser=lookup_parser)
    @api.response(301)
    def get(self, title, session=None):
        kwargs = lookup_parser.parse_args()
        kwargs['title'] = title
        return redirect(api.url_for(TraktSeriesSearchApi, **kwargs), code=301)


@trakt_api.route('/movies/')
class TraktMovieSearchApi(APIResource):
    @etag(cache_age=3600)
    @api.response(200, 'Successfully found show', movie_return_schema)
    @api.response(NotFoundError)
    @api.doc(parser=lookup_parser)
    def get(self, session=None):
        """Trakt movie lookup"""
        args = lookup_parser.parse_args()
        include_actors = args.pop('include_actors')
        include_translations = args.pop('include_translations')
        kwargs = args
        try:
            movie = ApiTrakt.lookup_movie(session=session, **kwargs)
        except LookupError as e:
            raise NotFoundError(e.args[0])
        result = movie.to_dict()
        if include_actors:
            result['actors'] = db.list_actors(movie.actors)
        if include_translations:
            result['translations'] = db.get_translations_dict(movie.translations, 'movie')
        return jsonify(result)


@trakt_api.route('/movies/<string:title>')
@api.doc(params={'title': 'Movie name'})
class TraktMovieWithTitleSearchApi(APIResource):
    @api.doc(parser=lookup_parser)
    @api.response(301)
    def get(self, title, session=None):
        kwargs = lookup_parser.parse_args()
        kwargs['title'] = title
        return redirect(api.url_for(TraktMovieSearchApi, **kwargs), code=301)



import itertools

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils.template import get_template

logger = logger.bind(name='notify_entry')

ENTRY_CONTAINERS = ['entries', 'accepted', 'rejected', 'failed', 'undecided']

VIA_SCHEMA = {
    'type': 'array',
    'items': {
        'allOf': [
            {'$ref': '/schema/plugins?interface=notifiers'},
            {
                'maxProperties': 1,
                'error_maxProperties': 'Plugin options indented 2 more spaces than '
                'the first letter of the plugin name.',
                'minProperties': 1,
            },
        ]
    },
}


class Notify:
    schema = {
        'type': 'object',
        'properties': {
            'entries': {
                'type': 'object',
                'properties': {
                    'title': {'type': 'string', 'default': '{{ title }}'},
                    'message': {
                        'type': 'string',
                        'default': '{% if series_name is defined %}'
                        '{{ tvdb_series_name|d(series_name) }} '
                        '{{series_id}} {{tvdb_ep_name|d('
                        ')}}'
                        '{% elif imdb_name is defined %}'
                        '{{imdb_name}} {{imdb_year}}'
                        '{% elif title is defined %}'
                        '{{ title }}'
                        '{% endif %}',
                    },
                    'template': {'type': 'string'},
                    'what': one_or_more({'type': 'string', 'enum': ENTRY_CONTAINERS}),
                    'via': VIA_SCHEMA,
                },
                'required': ['via'],
                'additionalProperties': False,
            },
            'task': {
                'type': 'object',
                'properties': {
                    'title': {
                        'type': 'string',
                        'default': '{% if not task.failed and not task.accepted %} Task {{task.name}} did not'
                        ' produce any entries.'
                        '{% else %} [FlexGet] {{task.name}}:'
                        '{% if task.failed %} {{task.failed|length}} failed entries.{% endif %}'
                        '{% if task.accepted %} {{task.accepted|length}} new entries downloaded.{% endif %}'
                        '{% endif %}',
                    },
                    'message': {'type': 'string'},
                    'template': {'type': 'string', 'default': 'default.template'},
                    'always_send': {'type': 'boolean', 'default': False},
                    'via': VIA_SCHEMA,
                },
                'required': ['via'],
                'additionalProperties': False,
            },
            'abort': {
                'type': 'object',
                'properties': {
                    'title': {'type': 'string', 'default': 'Task {{ task.name }} has aborted!'},
                    'message': {'type': 'string', 'default': 'Reason: {{ task.abort_reason }}'},
                    'via': VIA_SCHEMA,
                },
                'required': ['via'],
            },
        },
        'additionalProperties': False,
        'minProperties': 1,
        'error_minProperties': 'You must specify at least one of `entries` or `task` in your notify config.',
    }

    def prepare_config(self, config):
        if 'entries' in config:
            config['entries'].setdefault('what', ['accepted'])
            if not isinstance(config['entries']['what'], list):
                config['entries']['what'] = [config['entries']['what']]
        return config

    def send_notification(self, *args, **kwargs):
        send_notification = plugin.get('notification_framework', 'notify').send_notification
        try:
            send_notification(*args, **kwargs)
        except plugin.PluginError as e:
            logger.error(e)
        except plugin.PluginWarning as e:
            logger.warning(e)
        except Exception as e:
            logger.exception(e)

    @plugin.priority(0)
    def on_task_output(self, task, config):
        config = self.prepare_config(config)
        if 'entries' in config:
            entries = list(
                itertools.chain(*(getattr(task, what) for what in config['entries']['what']))
            )
            if not entries:
                logger.debug('No entries to notify about.')
            else:
                # If a file template is defined, it overrides message
                if config['entries'].get('template'):
                    try:
                        message = get_template(config['entries']['template'], scope='entry')
                    except ValueError:
                        raise plugin.PluginError(
                            'Cannot locate template on disk: %s' % config['entries']['template']
                        )
                else:
                    message = config['entries']['message']
                for entry in entries:
                    self.send_notification(
                        config['entries']['title'],
                        message,
                        config['entries']['via'],
                        template_renderer=entry.render,
                    )
        if 'task' in config:
            if not (task.accepted or task.failed) and not config['task']['always_send']:
                logger.verbose('No accepted or failed entries, not sending a notification.')
                return
            if config['task'].get('message'):
                template = config['task']['message']
            else:
                try:
                    template = get_template(config['task']['template'], scope='task')
                except ValueError:
                    raise plugin.PluginError(
                        'Cannot locate template on disk: %s' % config['task']['template']
                    )
            self.send_notification(
                config['task']['title'],
                template,
                config['task']['via'],
                template_renderer=task.render,
            )

    def on_task_abort(self, task, config):
        if 'abort' in config:
            if task.silent_abort:
                return
            logger.debug('sending abort notification')
            self.send_notification(
                config['abort']['title'],
                config['abort']['message'],
                config['abort']['via'],
                template_renderer=task.render,
            )


@event('plugin.register')
def register_plugin():
    plugin.register(Notify, 'notify', api_ver=2)



from flask import jsonify
from flask_restx import inputs

from flexget.api import APIResource, api
from flexget.api.app import BadRequest, NotFoundError, etag
from flexget.components.thetvdb.api_tvdb import lookup_episode, lookup_series, search_for_series

tvdb_api = api.namespace('tvdb', description='TheTVDB Shows')


class ObjectsContainer:
    tvdb_series_object = {
        'type': 'object',
        'properties': {
            'tvdb_id': {'type': 'integer'},
            'last_updated': {'type': 'string', 'format': 'date-time'},
            'expired': {'type': 'boolean'},
            'series_name': {'type': 'string'},
            'rating': {'type': 'number'},
            'status': {'type': 'string'},
            'runtime': {'type': 'integer'},
            'airs_time': {'type': 'string'},
            'airs_dayofweek': {'type': 'string'},
            'content_rating': {'type': 'string'},
            'network': {'type': 'string'},
            'overview': {'type': 'string'},
            'imdb_id': {'type': 'string'},
            'zap2it_id': {'type': 'string'},
            'banner': {'type': 'string'},
            'first_aired': {'type': 'string'},
            'actors': {'type': 'array', 'items': {'type': 'string'}},
            'aliases': {'type': 'array', 'items': {'type': 'string'}},
            'posters': {'type': 'array', 'items': {'type': 'string'}},
            'genres': {'type': 'array', 'items': {'type': 'string'}},
            'language': {'type': 'string'},
        },
        'required': [
            'tvdb_id',
            'last_updated',
            'expired',
            'series_name',
            'rating',
            'status',
            'runtime',
            'airs_time',
            'airs_dayofweek',
            'content_rating',
            'network',
            'overview',
            'imdb_id',
            'zap2it_id',
            'banner',
            'first_aired',
            'aliases',
            'posters',
            'genres',
            'language',
        ],
        'additionalProperties': False,
    }

    episode_object = {
        'type': 'object',
        'properties': {
            'id': {'type': 'integer'},
            'expired': {'type': 'boolean'},
            'last_update': {'type': 'integer'},
            'season_number': {'type': 'integer'},
            'episode_number': {'type': 'integer'},
            'absolute_number': {'type': ['integer', 'null']},
            'episode_name': {'type': 'string'},
            'overview': {'type': 'string'},
            'director': {'type': 'string'},
            'rating': {'type': 'number'},
            'image': {'type': ['string', 'null']},
            'first_aired': {'type': 'string'},
            'series_id': {'type': 'integer'},
        },
        'required': [
            'id',
            'expired',
            'last_update',
            'season_number',
            'episode_number',
            'absolute_number',
            'episode_name',
            'overview',
            'director',
            'rating',
            'image',
            'first_aired',
            'series_id',
        ],
        'additionalProperties': False,
    }

    search_result_object = {
        'type': 'object',
        'properties': {
            'aliases': {'type': 'array', 'items': {'type': 'string'}},
            'first_aired': {'type': 'string', 'format': 'date-time'},
            'banner': {'type': ['string', 'null']},
            'network': {'type': 'string'},
            'series_name': {'type': 'string'},
            'status': {'type': 'string'},
            'overview': {'type': ['string', 'null']},
            'tvdb_id': {'type': 'integer'},
        },
        'required': [
            'aliases',
            'first_aired',
            'banner',
            'network',
            'series_name',
            'status',
            'overview',
            'tvdb_id',
        ],
        'additionalProperties': False,
    }
    search_results_object = {'type': 'array', 'items': search_result_object}


tvdb_series_schema = api.schema_model('tvdb_series_schema', ObjectsContainer.tvdb_series_object)
tvdb_episode_schema = api.schema_model('tvdb_episode_schema', ObjectsContainer.episode_object)
search_results_schema = api.schema_model(
    'tvdb_search_results_schema', ObjectsContainer.search_results_object
)

base_parser = api.parser()
base_parser.add_argument(
    'language', default='en', help='Language abbreviation string for different language support'
)

series_parser = base_parser.copy()
series_parser.add_argument(
    'include_actors', type=inputs.boolean, help='Include actors in response'
)


@tvdb_api.route('/series/<string:title>/')
@api.doc(params={'title': 'TV Show name or TVDB ID'}, parser=series_parser)
class TVDBSeriesLookupAPI(APIResource):
    @etag(cache_age=3600)
    @api.response(200, 'Successfully found show', tvdb_series_schema)
    @api.response(NotFoundError)
    def get(self, title, session=None):
        """TheTVDB series lookup"""
        args = series_parser.parse_args()
        language = args['language']

        try:
            tvdb_id = int(title)
        except ValueError:
            tvdb_id = None

        kwargs = {'session': session, 'language': language}

        if tvdb_id:
            kwargs['tvdb_id'] = tvdb_id
        else:
            kwargs['name'] = title

        try:
            series = lookup_series(**kwargs)
        except LookupError as e:
            raise NotFoundError(e.args[0])

        result = series.to_dict()
        if args.get('include_actors'):
            result['actors'] = series.actors
        return jsonify(result)


episode_parser = base_parser.copy()
episode_parser.add_argument('season_number', type=int, help='Season number')
episode_parser.add_argument('ep_number', type=int, help='Episode number')
episode_parser.add_argument('absolute_number', type=int, help='Absolute episode number')
episode_parser.add_argument(
    'air_date', type=inputs.date, help='Episode airdate in `YYYY-mm-dd` format'
)


@tvdb_api.route('/episode/<int:tvdb_id>/')
@api.doc(params={'tvdb_id': 'TVDB ID of show'}, parser=episode_parser)
class TVDBEpisodeSearchAPI(APIResource):
    @etag(cache_age=3600)
    @api.response(200, 'Successfully found episode', tvdb_episode_schema)
    @api.response(NotFoundError)
    @api.response(BadRequest)
    def get(self, tvdb_id, session=None):
        """TheTVDB episode lookup"""
        args = episode_parser.parse_args()
        language = args['language']

        absolute_number = args.get('absolute_number')
        season_number = args.get('season_number')
        ep_number = args.get('ep_number')
        air_date = args.get('air_date')

        if not ((season_number and ep_number) or absolute_number or air_date):
            raise BadRequest(
                'not enough parameters for lookup. Either season and episode number or absolute number '
                'are required.'
            )
        kwargs = {'tvdb_id': tvdb_id, 'session': session, 'language': language}

        if absolute_number:
            kwargs['absolute_number'] = absolute_number
        if season_number and ep_number:
            kwargs['season_number'] = season_number
            kwargs['episode_number'] = ep_number
        if air_date:
            kwargs['first_aired'] = air_date

        try:
            episode = lookup_episode(**kwargs)
        except LookupError as e:
            raise NotFoundError(e.args[0])
        return jsonify(episode.to_dict())


search_parser = base_parser.copy()
search_parser.add_argument('search_name', help='Series Name')
search_parser.add_argument('imdb_id', help='Series IMDB ID')
search_parser.add_argument('zap2it_id', help='Series ZAP2IT ID')
search_parser.add_argument(
    'force_search',
    type=inputs.boolean,
    help='Force online lookup or allow for result to be retrieved from cache',
)


@tvdb_api.route('/search/')
@api.doc(parser=search_parser)
class TVDBSeriesSearchAPI(APIResource):
    @etag(cache_age=3600)
    @api.response(200, 'Successfully got results', search_results_schema)
    @api.response(BadRequest)
    @api.response(NotFoundError)
    def get(self, session=None):
        """TheTVDB series search"""
        args = search_parser.parse_args()
        language = args['language']

        search_name = args.get('search_name')
        imdb_id = args.get('imdb_id')
        zap2it_id = args.get('zap2it_id')
        force_search = args.get('force_search')

        if not any(arg for arg in [search_name, imdb_id, zap2it_id]):
            raise BadRequest('Not enough lookup arguments')
        kwargs = {
            'search_name': search_name,
            'imdb_id': imdb_id,
            'zap2it_id': zap2it_id,
            'force_search': force_search,
            'session': session,
            'language': language,
        }
        try:
            search_results = search_for_series(**kwargs)
        except LookupError as e:
            raise NotFoundError(e.args[0])
        return jsonify([a.to_dict() for a in search_results])



import binascii
import itertools
import socket
import struct
import threading
from http.client import BadStatusLine
from random import randrange
from urllib.error import URLError
from urllib.parse import quote, urlparse, urlsplit, urlunsplit

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.event import event
from flexget.utils import requests
from flexget.utils.bittorrent import bdecode

logger = logger.bind(name='torrent_alive')


class TorrentAliveThread(threading.Thread):
    _counter = itertools.count()

    def __init__(self, tracker, info_hash):
        threading.Thread.__init__(self, name='torrent_alive-%d' % next(self._counter))
        self.tracker = tracker
        self.info_hash = info_hash
        self.tracker_seeds = 0

    def run(self):
        self.tracker_seeds = get_tracker_seeds(self.tracker, self.info_hash)
        logger.debug(
            '{} seeds found from {}',
            self.tracker_seeds,
            get_scrape_url(self.tracker, self.info_hash),
        )


def max_seeds_from_threads(threads):
    """
    Joins the threads and returns the maximum seeds found from any of them.

    :param threads: A list of started `TorrentAliveThread`
    :return: Maximum seeds found from any of the threads
    """
    seeds = 0
    for background in threads:
        logger.debug('Coming up next: {}', background.tracker)
        background.join()
        seeds = max(seeds, background.tracker_seeds)
        logger.debug('Current highest number of seeds found: {}', seeds)
    return seeds


def get_scrape_url(tracker_url, info_hash):
    if 'announce' in tracker_url:
        v = urlsplit(tracker_url)
        result = urlunsplit(
            [v.scheme, v.netloc, v.path.replace('announce', 'scrape'), v.query, v.fragment]
        )
    else:
        logger.debug('`announce` not contained in tracker url, guessing scrape address.')
        result = tracker_url + '/scrape'

    result += '&' if '?' in result else '?'
    result += 'info_hash=%s' % quote(binascii.unhexlify(info_hash))
    return result


def get_udp_seeds(url, info_hash):
    parsed_url = urlparse(url)
    try:
        port = parsed_url.port
    except ValueError:
        logger.error('UDP Port Error, url was {}', url)
        return 0

    logger.debug('Checking for seeds from {}', url)

    connection_id = 0x41727101980  # connection id is always this
    transaction_id = randrange(1, 65535)  # Random Transaction ID creation

    if port is None:
        logger.error('UDP Port Error, port was None')
        return 0

    if port < 0 or port > 65535:
        logger.error('UDP Port Error, port was {}', port)
        return 0

    # Create the socket
    try:
        clisocket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        clisocket.settimeout(5.0)
        clisocket.connect((parsed_url.hostname, port))

        # build packet with connection_ID, using 0 value for action, giving our transaction ID for this packet
        packet = struct.pack(b">QLL", connection_id, 0, transaction_id)
        clisocket.send(packet)

        # set 16 bytes ["QLL" = 16 bytes] for the fmq for unpack
        res = clisocket.recv(16)
        # check recieved packet for response
        action, transaction_id, connection_id = struct.unpack(b">LLQ", res)

        # build packet hash out of decoded info_hash
        packet_hash = binascii.unhexlify(info_hash)

        # construct packet for scrape with decoded info_hash setting action byte to 2 for scape
        packet = struct.pack(b">QLL", connection_id, 2, transaction_id) + packet_hash

        clisocket.send(packet)
        # set recieve size of 8 + 12 bytes
        res = clisocket.recv(20)

    except OSError as e:
        logger.warning('Socket Error: {}', e)
        return 0
    # Check for UDP error packet
    (action,) = struct.unpack(b">L", res[:4])
    if action == 3:
        logger.error('There was a UDP Packet Error 3')
        return 0

    # first 8 bytes are followed by seeders, completed and leechers for requested torrent
    seeders, _, _ = struct.unpack(b">LLL", res[8:20])
    logger.debug('get_udp_seeds is returning: {}', seeders)
    clisocket.close()
    return seeders


def get_http_seeds(url, info_hash):
    url = get_scrape_url(url, info_hash)
    if not url:
        logger.debug('if not url is true returning 0')
        return 0
    logger.debug('Checking for seeds from {}', url)

    try:
        data = bdecode(requests.get(url).content).get('files')
    except RequestException as e:
        logger.debug('Error scraping: {}', e)
        return 0
    except SyntaxError as e:
        logger.warning('Error decoding tracker response: {}', e)
        return 0
    except BadStatusLine as e:
        logger.warning('Error BadStatusLine: {}', e)
        return 0
    except OSError as e:
        logger.warning('Server error: {}', e)
        return 0
    if not data:
        logger.debug('No data received from tracker scrape.')
        return 0
    logger.debug('get_http_seeds is returning: {}', list(data.values())[0]['complete'])
    return list(data.values())[0]['complete']


def get_tracker_seeds(url, info_hash):
    if url.startswith('udp'):
        return get_udp_seeds(url, info_hash)
    elif url.startswith('http'):
        return get_http_seeds(url, info_hash)
    else:
        logger.warning('There is a problem with the get_tracker_seeds')
        return 0


class TorrentAlive:
    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'integer'},
            {
                'type': 'object',
                'properties': {
                    'min_seeds': {'type': 'integer'},
                    'reject_for': {'type': 'string', 'format': 'interval'},
                },
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        # Convert config to dict form
        if not isinstance(config, dict):
            config = {'min_seeds': int(config)}
        # Set the defaults
        config.setdefault('min_seeds', 1)
        config.setdefault('reject_for', '1 hour')
        return config

    @plugin.priority(150)
    def on_task_filter(self, task, config):
        if not config:
            return
        config = self.prepare_config(config)
        for entry in task.entries:
            if 'torrent_seeds' in entry and entry['torrent_seeds'] < config['min_seeds']:
                entry.reject(
                    reason='Had < %d required seeds. (%s)'
                    % (config['min_seeds'], entry['torrent_seeds'])
                )

    # Run on output phase so that we let torrent plugin output modified torrent file first
    @plugin.priority(250)
    def on_task_output(self, task, config):
        if not config:
            return
        config = self.prepare_config(config)
        min_seeds = config['min_seeds']

        for entry in task.accepted:
            # If torrent_seeds is filled, we will have already filtered in filter phase
            if entry.get('torrent_seeds'):
                logger.debug(
                    'Not checking trackers for seeds, as torrent_seeds is already filled.'
                )
                continue
            logger.debug('Checking for seeds for {}:', entry['title'])
            torrent = entry.get('torrent')
            if torrent:
                logger.debug('started examining torrent: {}', torrent)
                seeds = 0
                info_hash = torrent.info_hash
                announce_list = torrent.content.get('announce-list')
                if announce_list:
                    # Multitracker torrent
                    threadlist = []
                    for tier in announce_list:
                        for tracker in tier:
                            background = TorrentAliveThread(tracker, info_hash)
                            try:
                                background.start()
                                threadlist.append(background)
                            except threading.ThreadError:
                                # If we can't start a new thread, wait for current ones to complete and continue
                                logger.debug('Reached max threads, finishing current threads.')
                                seeds = max(seeds, max_seeds_from_threads(threadlist))
                                background.start()
                                threadlist = [background]
                            logger.debug(
                                'Started thread to scrape {} with info hash {}', tracker, info_hash
                            )

                    seeds = max(seeds, max_seeds_from_threads(threadlist))
                    logger.debug('Highest number of seeds found: {}', seeds)
                elif torrent.content.get('announce'):
                    # Single tracker
                    tracker = torrent.content['announce']
                    try:
                        seeds = get_tracker_seeds(tracker, info_hash)
                    except URLError as e:
                        logger.debug('Error scraping {}: {}', tracker, e)
                else:
                    logger.warning(
                        'Torrent {} does not seem to have a tracker specified, cannot check for seeders',
                        entry['title'],
                    )
                    return

                # Reject if needed
                if seeds < min_seeds:
                    entry.reject(
                        reason='Tracker(s) had < %s required seeds. (%s)' % (min_seeds, seeds),
                        remember_time=config['reject_for'],
                    )
                    # Maybe there is better match that has enough seeds
                    task.rerun(plugin='torrent_alive', reason='Not enough seeds')
                else:
                    logger.debug('Found {} seeds from trackers', seeds)


@event('plugin.register')
def register_plugin():
    plugin.register(TorrentAlive, 'torrent_alive', api_ver=2)



from flask import jsonify

from flexget.api import APIResource, api
from flexget.api.app import etag
from flexget.components.imdb.utils import ImdbSearch

imdb_api = api.namespace('imdb', description='IMDB lookup endpoint')


class ObjectsContainer:
    movie_object = {
        'type': 'object',
        'properties': {
            'imdb_id': {'type': 'string'},
            'match': {'type': 'number'},
            'name': {'type': 'string'},
            'url': {'type': 'string'},
            'year': {'type': 'number'},
            'thumbnail': {'type': 'string'},
        },
        'required': ['imdb_id', 'match', 'name', 'url', 'year'],
        'additionalProperties': False,
    }

    return_object = {'type': 'array', 'items': movie_object}


return_schema = api.schema_model('imdb_search_schema', ObjectsContainer.return_object)


@imdb_api.route('/search/<string:title>/')
@api.doc(params={'title': 'Movie name or IMDB ID'})
class IMDBMovieSearch(APIResource):
    # noinspection PyUnusedLocal
    @etag
    @api.response(200, model=return_schema)
    def get(self, title, session=None):
        """ Get a list of IMDB search result by name or ID"""
        raw_movies = ImdbSearch().smart_match(title, single_match=False)
        if not raw_movies:
            return jsonify([])
        # Convert single movie to list to preserve consistent reply
        if not isinstance(raw_movies, list):
            raw_movies = [raw_movies]
        return jsonify(raw_movies)



#!/usr/bin/python
import os
import sys
from typing import Sequence

# __version__ import need to be first in order to avoid circular import within logger
from ._version import __version__  # noqa

# isort: split
from flexget import log  # noqa
from flexget.manager import Manager  # noqa


def main(args: Sequence = None):
    """Main entry point for Command Line Interface"""

    try:
        log.initialize()

        try:
            manager = Manager(args)
        except (OSError, ValueError) as e:
            if _is_debug():
                import traceback

                traceback.print_exc()
            else:
                print('Could not instantiate manager: %s' % e, file=sys.stderr)
            sys.exit(1)

        try:
            if manager.options.profile:
                try:
                    import cProfile as profile
                except ImportError:
                    import profile
                profile.runctx(
                    'manager.start()',
                    globals(),
                    locals(),
                    os.path.join(manager.config_base, manager.options.profile),
                )
            else:
                manager.start()
        except (OSError, ValueError) as e:
            if _is_debug():
                import traceback

                traceback.print_exc()
            else:
                print('Could not start manager: %s' % e, file=sys.stderr)

            sys.exit(1)
    except KeyboardInterrupt:
        if _is_debug():
            import traceback

            traceback.print_exc()

        print('Killed with keyboard interrupt.', file=sys.stderr)
        sys.exit(1)


def _is_debug() -> bool:
    return any(
        arg in ['debug', '--debug', '--loglevel=trace', '--loglevel=debug']
        for arg in [a.lower() for a in sys.argv]
    )



import datetime

from loguru import logger

from flexget import options, plugin
from flexget.config_schema import parse_interval
from flexget.event import event

logger = logger.bind(name='interval')


class PluginInterval:
    """
    Allows specifying minimum interval for task execution.

    Format: [n] [minutes|hours|days|weeks]

    Example:

    interval: 7 days
    """

    schema = {'type': 'string', 'format': 'interval'}

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_start(self, task, config):
        if task.options.learn:
            logger.info('Ignoring task {} interval for --learn', task.name)
            return
        last_time = task.simple_persistence.get('last_time')
        if not last_time:
            logger.info('No previous run recorded, running now')
        elif task.options.interval_ignore:
            logger.info('Ignoring interval because of --now')
        else:
            logger.debug('last_time: {!r}', last_time)
            logger.debug('interval: {}', config)
            next_time = last_time + parse_interval(config)
            logger.debug('next_time: {!r}', next_time)
            if datetime.datetime.now() < next_time:
                logger.verbose(
                    'Interval {} not met on task {}. Use --now to override.', config, task.name
                )
                task.abort('Interval not met', silent=True)
                return
        logger.debug('interval passed')
        task.simple_persistence['last_time'] = datetime.datetime.now()


@event('plugin.register')
def register_plugin():
    plugin.register(PluginInterval, 'interval', api_ver=2)


@event('options.register')
def register_parser_arguments():
    options.get_parser('execute').add_argument(
        '--now',
        action='store_true',
        dest='interval_ignore',
        default=False,
        help='run task(s) even if the interval plugin would normally prevent it',
    )



# -*- coding: utf-8 -*-


import re

import feedparser
from urllib.request import ProxyHandler
from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.plugin import PluginError
from flexget.utils.requests import RequestException
from flexget.utils.soup import get_soup

__author__ = 'danfocus'

logger = logger.bind(name='lostfilm')

EPISODE_REGEXP = re.compile(r'.*/series/.*/season_(\d+)/episode_(\d+)/.*')
LOSTFILM_ID_REGEXP = re.compile(r'.*static\.lostfilm\..*/Images/(\d+)/Posters/.*')
TEXT_REGEXP = re.compile(r'^\d+\s+сезон\s+\d+\s+серия\.\s(.+)\s\((.+)\)$')

quality_map = {
    'SD': ('480p', 'avi'),
    '1080': ('1080p', 'mkv'),
    'MP4': ('720p', 'mp4'),
    'HD': ('720p', 'mkv'),
}

LOSTFILM_URL = 'https://lostfilm.tv/rss.xml'


class LostFilm:
    """
    Change new lostfilm's rss links

    Example::

      lostfilm: yes

    Advanced usage:

      lostfilm:
        url: <url>
    """

    schema = {
        'type': ['boolean', 'object'],
        'properties': {'url': {'type': 'string', 'format': 'url'}},
        'additionalProperties': False,
    }

    def build_config(self, config):
        """Set default url to config"""
        if isinstance(config, bool) and config is True:
            config = {'url': LOSTFILM_URL}
        return config

    def on_task_input(self, task, config):
        config = self.build_config(config)
        if config is False:
            return
        proxy_handler = None
        if task.requests.proxies is not None:
            proxy_handler = ProxyHandler(task.requests.proxies)
        try:
            rss = feedparser.parse(config['url'], handlers=[proxy_handler])
        except Exception:
            raise PluginError('Cannot parse rss feed')
        status = rss.get('status')
        if status != 200:
            raise PluginError('Received %s status instead of 200 (OK)' % status)
        entries = []
        for item in rss.entries:
            if item.get('link') is None:
                logger.debug('Item doesn\'t have a link')
                continue
            if item.get('description') is None:
                logger.debug('Item doesn\'t have a description')
                continue
            try:
                lostfilm_num = LOSTFILM_ID_REGEXP.search(item['description']).groups()
            except Exception:
                logger.debug('Item doesn\'t have lostfilm id in description')
                continue
            try:
                season_num, episode_num = [
                    int(x) for x in EPISODE_REGEXP.search(item['link']).groups()
                ]
            except Exception:
                logger.debug('Item doesn\'t have episode id in link')
                continue
            params = {'c': lostfilm_num, 's': season_num, 'e': episode_num}
            redirect_url = 'https://www.lostfilm.tv/v_search.php'
            try:
                response = task.requests.get(redirect_url, params=params)
            except RequestException as e:
                logger.error('Could not connect to redirect url: {:s}'.format(e))
                continue

            page = get_soup(response.content)
            try:
                redirect_url = page.head.meta['content'].split('url=')[1]
            except Exception:
                logger.error('Missing redirect')
                continue

            try:
                response = task.requests.get(redirect_url)
            except RequestException as e:
                logger.error('Could not connect to redirect url2: {:s}', e)
                continue

            page = get_soup(response.content)

            episode_name_rus = episode_name_eng = series_name_rus = None
            series_name_eng = None
            try:
                series_name_rus = page.find('div', 'inner-box--title').text.strip()
                title_eng_div = page.find('div', 'inner-box--subtitle').text.strip() or None
                series_name_eng = (
                    (title_eng_div.endswith(', сериал')) and title_eng_div[:-8] or None
                )
                text_div = page.find('div', 'inner-box--text').text.strip() or None
                episode_name_rus, episode_name_eng = TEXT_REGEXP.findall(text_div).pop()
                episode_name_rus = episode_name_rus.strip()
                episode_name_eng = episode_name_eng.strip()
            except Exception:
                logger.debug('Cannot parse head info')
                continue

            episode_id = 'S{:02d}E{:02d}'.format(season_num, episode_num)

            for item in page.findAll('div', 'inner-box--item'):
                torrent_link = quality = None
                try:
                    torrent_link = item.find('div', 'inner-box--link sub').a['href']
                    quality = item.find('div', 'inner-box--label').text.strip()
                except Exception:
                    logger.debug('Item doesn\'t have a link or quality')
                    continue
                if torrent_link is None or quality is None:
                    logger.debug('Item doesn\'t have a link or quality')
                    continue
                if quality_map.get(quality):
                    quality, file_ext = quality_map.get(quality)
                else:
                    file_ext = 'avi'
                if series_name_eng:
                    new_title = '.'.join(
                        [
                            series_name_eng,
                            episode_id,
                            quality,
                            'rus.LostFilm.TV',
                            file_ext,
                            'torrent',
                        ]
                    ).replace(' ', '.')
                else:
                    if item.get('title') is not None:
                        new_title = '{} {}'.format(item['title'], quality)
                    else:
                        logger.debug('Item doesn\'t have a title')
                        continue
                new_entry = Entry()
                new_entry['url'] = torrent_link
                new_entry['title'] = new_title.strip()
                if series_name_rus:
                    new_entry['series_name_rus'] = series_name_rus
                if episode_name_rus:
                    new_entry['episode_name_rus'] = episode_name_rus
                if series_name_eng:
                    new_entry['series_name_eng'] = series_name_eng
                if episode_name_eng:
                    new_entry['episode_name_eng'] = episode_name_eng
                entries.append(new_entry)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(LostFilm, 'lostfilm', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='headers')


class PluginHeaders:
    """Allow setting up any headers in all requests (which use urllib2)

    Example:

    headers:
      cookie: uid=<YOUR UID>; pass=<YOUR PASS>
    """

    schema = {'type': 'object', 'additionalProperties': {'type': 'string'}}

    @plugin.priority(130)
    def on_task_start(self, task, config):
        """Task starting"""
        # Set the headers for this task's request session
        logger.debug('headers to add: {}', config)
        if task.requests.headers:
            task.requests.headers.update(config)
        else:
            task.requests.headers = config


@event('plugin.register')
def register_plugin():
    plugin.register(PluginHeaders, 'headers', api_ver=2)



# coding=utf-8


import re

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.event import event
from flexget.utils import requests
from flexget.utils.soup import get_soup

logger = logger.bind(name='serienjunkies')

regex_single_ep = re.compile(r'(S\d+E\d\d+)(?!-E)', re.I)
regex_multi_ep = re.compile(r'(?P<season>S\d\d)E(?P<startep>\d\d+)-E?(?P<stopep>\d\d+)', re.I)
regex_season = re.compile(r'(?<=\.|\-)S\d\d(?:[-\.]S\d\d)*(?!E\d\d+)', re.I)

regex_language_container = re.compile(r'Sprache')

regex_is_german = re.compile(r'german|deutsch', re.I)
regex_is_foreign = re.compile(
    r'englisc?h|französisch|japanisch|dänisch|norwegisch|niederländisch|ungarisch|italienisch|portugiesisch',
    re.I,
)
regex_is_subtitle = re.compile(r'Untertitel|Subs?|UT', re.I)

LANGUAGE = ['german', 'foreign', 'subtitle', 'dual']
HOSTER = ['ul', 'cz', 'so', 'all']

DEFAULT_LANGUAGE = 'dual'
DEFAULT_HOSTER = 'ul'


class UrlRewriteSerienjunkies:
    """
    Serienjunkies urlrewriter
    Version 1.0.2

    Language setting works like a whitelist, the selected is needed,
    but others are still possible.

    Configuration
    language: [german|foreign|subtitle|dual] default "foreign"
    hoster: [ul|cz|so|all] default "ul"
    """

    schema = {
        'type': 'object',
        'properties': {
            'language': {'type': 'string', 'enum': LANGUAGE},
            'hoster': {'type': 'string', 'enum': HOSTER},
        },
        'additionalProperties': False,
    }

    # Since the urlrewriter relies on a config, we need to create a default one
    config = {'hoster': DEFAULT_HOSTER, 'language': DEFAULT_LANGUAGE}

    def on_task_start(self, task, config):
        self.config = config

    # urlrewriter API
    def url_rewritable(self, task, entry):
        url = entry['url']
        if url.startswith('http://www.serienjunkies.org/') or url.startswith(
            'http://serienjunkies.org/'
        ):
            return True
        return False

    # urlrewriter API
    def url_rewrite(self, task, entry):
        series_url = entry['url']
        search_title = re.sub(r'\[.*\] ', '', entry['title'])

        download_urls = self.parse_downloads(series_url, search_title)
        if not download_urls:
            entry.reject('No Episode found')
        else:
            entry['url'] = download_urls[-1]
            entry['description'] = ", ".join(download_urls)

        # Debug Information
        logger.debug('TV Show URL: {}', series_url)
        logger.debug('Episode: {}', search_title)
        logger.debug('Download URL: {}', download_urls)

    @plugin.internet(logger)
    def parse_downloads(self, series_url, search_title):
        page = requests.get(series_url).content
        try:
            soup = get_soup(page)
        except Exception as e:
            raise UrlRewritingError(e)

        urls = []
        # find all titles
        episode_titles = self.find_all_titles(search_title)
        if not episode_titles:
            raise UrlRewritingError('Unable to find episode')

        for ep_title in episode_titles:
            # find matching download
            episode_title = soup.find('strong', text=re.compile(ep_title, re.I))
            if not episode_title:
                continue

            # find download container
            episode = episode_title.parent
            if not episode:
                continue

            # find episode language
            episode_lang = episode.find_previous('strong', text=re.compile('Sprache')).next_sibling
            if not episode_lang:
                logger.warning('No language found for: {}', series_url)
                continue

            # filter language
            if not self.check_language(episode_lang):
                logger.warning(
                    'languages not matching: {} <> {}', self.config['language'], episode_lang
                )
                continue

            # find download links
            links = episode.find_all('a')
            if not links:
                logger.warning('No links found for: {}', series_url)
                continue

            for link in links:
                if not link.has_attr('href'):
                    continue

                url = link['href']
                pattern = (
                    r'http:\/\/download\.serienjunkies\.org.*%s_.*\.html' % self.config['hoster']
                )

                if re.match(pattern, url) or self.config['hoster'] == 'all':
                    urls.append(url)
                else:
                    continue
        return urls

    def find_all_titles(self, search_title):
        search_titles = []
        # Check type
        if regex_multi_ep.search(search_title):
            logger.debug('Title seems to describe multiple episodes')
            first_ep = int(regex_multi_ep.search(search_title).group('startep'))
            last_ep = int(regex_multi_ep.search(search_title).group('stopep'))
            season = regex_multi_ep.search(search_title).group('season') + 'E'
            for i in range(first_ep, last_ep + 1):
                # ToDO: Umlaute , Mehrzeilig etc.
                search_titles.append(
                    regex_multi_ep.sub(
                        season + str(i).zfill(2) + '[\\\\w\\\\.\\\\(\\\\)]*', search_title
                    )
                )
        elif regex_season.search(search_title):
            logger.debug('Title seems to describe one or more season')
            search_string = regex_season.search(search_title).group(0)
            for s in re.findall(r'(?<!\-)S\d\d(?!\-)', search_string):
                search_titles.append(regex_season.sub(s + '[\\\\w\\\\.]*', search_title))
            for s in re.finditer(r'(?<!\-)S(\d\d)-S(\d\d)(?!\-)', search_string):
                season_start = int(s.group(1))
                season_end = int(s.group(2))
                for i in range(season_start, season_end + 1):
                    search_titles.append(
                        regex_season.sub('S' + str(i).zfill(2) + '[\\\\w\\\\.]*', search_title)
                    )
        else:
            logger.debug('Title seems to describe a single episode')
            search_titles.append(re.escape(search_title))
        return search_titles

    def check_language(self, languages):
        # Cut additional Subtitles
        languages = languages.split('|', 1)[0]

        language_list = re.split(r'[,&]', languages)

        try:
            if self.config['language'] == 'german':
                if regex_is_german.search(language_list[0]):
                    return True
            elif self.config['language'] == 'foreign':
                if (regex_is_foreign.search(language_list[0]) and len(language_list) == 1) or (
                    len(language_list) > 1 and not regex_is_subtitle.search(language_list[1])
                ):
                    return True
            elif self.config['language'] == 'subtitle':
                if len(language_list) > 1 and regex_is_subtitle.search(language_list[1]):
                    return True
            elif self.config['language'] == 'dual':
                if len(language_list) > 1 and not regex_is_subtitle.search(language_list[1]):
                    return True
        except (KeyError, re.error):
            pass

        return False


@event('plugin.register')
def register_plugin():
    plugin.register(
        UrlRewriteSerienjunkies, 'serienjunkies', interfaces=['urlrewriter', 'task'], api_ver=2
    )



# coding=utf-8


import json
import re
from datetime import datetime, timedelta
from time import sleep

from loguru import logger
from requests.auth import AuthBase
from requests.exceptions import RequestException
from requests.utils import dict_from_cookiejar
from sqlalchemy import Column, DateTime, Integer, Unicode
from sqlalchemy.types import VARCHAR, TypeDecorator

from flexget import plugin
from flexget.db_schema import versioned_base
from flexget.event import event
from flexget.manager import Session
from flexget.plugin import PluginError

__author__ = 'asm0dey'

logger = logger.bind(name='rutracker_auth')
Base = versioned_base('rutracker_auth', 0)

MIRRORS = ['https://rutracker.net', 'https://rutracker.org']


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage:

        JSONEncodedDict(255)

    """

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)

        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class RutrackerAccount(Base):
    __tablename__ = 'rutracker_accoounts'
    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    login = Column(Unicode, index=True)
    cookies = Column(JSONEncodedDict)
    expiry_time = Column(DateTime)


class RutrackerAuth(AuthBase):
    """Supports downloading of torrents from 'rutracker' tracker
    if you pass cookies (CookieJar) to constructor then authentication will be bypassed and cookies will be just set
    """

    def update_base_url(self):
        url = None
        for mirror in MIRRORS:
            try:
                response = self.requests.get(mirror, timeout=2)
                if response.ok:
                    url = mirror
                    break
            except RequestException as err:
                logger.debug('Connection error. {}', str(err))

        if url:
            return url
        else:
            raise PluginError('Host unreachable.')

    def try_authenticate(self, payload):
        for _ in range(5):
            self.requests.post('{}/forum/login.php'.format(self.base_url), data=payload)
            if self.requests.cookies and len(self.requests.cookies) > 0:
                return self.requests.cookies
            else:
                sleep(3)
        raise PluginError('unable to obtain cookies from rutracker')

    def __init__(self, requests, login, password, cookies=None, db_session=None):
        self.requests = requests
        self.base_url = self.update_base_url()
        if cookies is None:
            logger.debug('rutracker cookie not found. Requesting new one')
            payload_ = {'login_username': login, 'login_password': password, 'login': 'Вход'}
            self.cookies_ = self.try_authenticate(payload_)
            if db_session:
                db_session.add(
                    RutrackerAccount(
                        login=login,
                        cookies=dict_from_cookiejar(self.cookies_),
                        expiry_time=datetime.now() + timedelta(days=1),
                    )
                )
                db_session.commit()
            else:
                raise ValueError('db_session can not be None if cookies is None')
        else:
            logger.debug('Using previously saved cookie')
            self.cookies_ = cookies

    def __call__(self, r):
        url = r.url
        t_id = re.findall(r'\d+', url)[0]
        data = 't={}'.format(t_id)
        headers = {
            'referer': '{}/forum/viewtopic.php?t={}'.format(self.base_url, t_id),
            'Content-Type': 'application/x-www-form-urlencoded',
            't': t_id,
            'Origin': self.base_url,
            'Accept-Encoding': 'gzip,deflate,sdch',
        }
        r.prepare_body(data=data, files=None)
        r.prepare_method('POST')
        r.prepare_url(url='{}/forum/dl.php?t={}'.format(self.base_url, t_id), params=None)
        r.prepare_headers(headers)
        r.prepare_cookies(self.cookies_)
        return r


class RutrackerUrlrewrite:
    """Usage:

    rutracker_auth:
      username: 'username_here'
      password: 'password_here'
    """

    schema = {
        'type': 'object',
        'properties': {'username': {'type': 'string'}, 'password': {'type': 'string'}},
        'additionalProperties': False,
    }

    auth_cache = {}

    @plugin.priority(127)
    def on_task_urlrewrite(self, task, config):
        username = config['username']
        db_session = Session()
        cookies = self.try_find_cookie(db_session, username)
        if username not in self.auth_cache:
            auth_handler = RutrackerAuth(
                task.requests, username, config['password'], cookies, db_session
            )
            self.auth_cache[username] = auth_handler
        else:
            auth_handler = self.auth_cache[username]
        for entry in task.accepted:
            if re.match(r'https?:\/\/rutracker', entry['url']):
                entry['download_auth'] = auth_handler

    @staticmethod
    def try_find_cookie(db_session, username):
        account = (
            db_session.query(RutrackerAccount).filter(RutrackerAccount.login == username).first()
        )
        if account:
            if account.expiry_time < datetime.now():
                db_session.delete(account)
                db_session.commit()
                return None
            return account.cookies
        else:
            return None


@event('plugin.register')
def register_plugin():
    plugin.register(RutrackerUrlrewrite, 'rutracker_auth', api_ver=2)



import base64
import datetime
import hashlib
import os

from loguru import logger
from sqlalchemy import Column, DateTime, Integer, String, Unicode

from flexget import db_schema, plugin
from flexget.event import event
from flexget.utils.sqlalchemy_utils import table_add_column, table_columns
from flexget.utils.template import RenderError, get_template, render_from_entry

logger = logger.bind(name='make_rss')
Base = db_schema.versioned_base('make_rss', 0)

rss2gen = True
try:
    import PyRSS2Gen
except ImportError:
    rss2gen = False


@db_schema.upgrade('make_rss')
def upgrade(ver, session):
    if ver is None:
        columns = table_columns('make_rss', session)
        if 'rsslink' not in columns:
            logger.info('Adding rsslink column to table make_rss.')
            table_add_column('make_rss', 'rsslink', String, session)
        ver = 0
    return ver


class RSSEntry(Base):
    __tablename__ = 'make_rss'

    id = Column(Integer, primary_key=True)
    title = Column(Unicode)
    description = Column(Unicode)
    link = Column(String)
    rsslink = Column(String)
    file = Column(Unicode)
    published = Column(DateTime, default=datetime.datetime.utcnow)
    enc_length = Column(Integer)
    enc_type = Column(String)


class OutputRSS:
    """
    Write RSS containing succeeded (downloaded) entries.

    Example::

      make_rss: ~/public_html/flexget.rss

    You may write into same file in multiple tasks.

    Example::

      my-task-A:
        make_rss: ~/public_html/series.rss
        .
        .
      my-task-B:
        make_rss: ~/public_html/series.rss
        .
        .

    With this example file series.rss would contain succeeded
    entries from both tasks.

    **Number of days / items**

    By default output contains items from last 7 days. You can specify
    different perioid, number of items or both. Value -1 means unlimited.

    Example::

      make_rss:
        file: ~/public_html/series.rss
        days: 2
        items: 10

    Generate RSS that will containing last two days and no more than 10 items.

    Example 2::

      make_rss:
        file: ~/public_html/series.rss
        days: -1
        items: 50

    Generate RSS that will contain last 50 items, regardless of dates.

    RSS feed properties:

    You can specify the URL, title, and description to include in tthe header
    of the RSS feed.

    Example::

      make_rss:
        file: ~/public_html/series.rss
        rsslink: http://my.server.net/series.rss
        rsstitle: The Flexget RSS Feed
        rssdesc: Episodes about Flexget.

    **RSS item title and link**

    You can specify the title and link for each item in the RSS feed.

    The item title can be any pattern that references fields in the input entry.

    The item link can be created from one of a list of fields in the input
    entry, in order of preference. The fields should be enumerated in a list.
    Note that the url field is always used as last possible fallback even
    without explicitly adding it into the list.

    Default field list for item URL: imdb_url, input_url, url

    Example::

      make_rss:
        file: ~/public_html/series.rss
        title: '{{title}} (from {{task}})'
        link:
          - imdb_url

    """

    schema = {
        'oneOf': [
            {'type': 'string'},  # TODO: path / file
            {
                'type': 'object',
                'properties': {
                    'file': {'type': 'string'},
                    'days': {'type': 'integer'},
                    'items': {'type': 'integer'},
                    'history': {'type': 'boolean'},
                    'timestamp': {'type': 'boolean'},
                    'rsslink': {'type': 'string'},
                    'rsstitle': {'type': 'string'},
                    'rssdesc': {'type': 'string'},
                    'encoding': {'type': 'string'},  # TODO: only valid choices
                    'title': {'type': 'string'},
                    'template': {'type': 'string'},
                    'link': {'type': 'array', 'items': {'type': 'string'}},
                },
                'required': ['file'],
                'additionalProperties': False,
            },
        ]
    }

    def on_task_output(self, task, config):
        # makes this plugin count as output (stops warnings about missing outputs)
        pass

    def prepare_config(self, config):
        if not isinstance(config, dict):
            config = {'file': config}
        config.setdefault('days', 7)
        config.setdefault('items', -1)
        config.setdefault('history', True)
        config.setdefault('encoding', 'UTF-8')
        config.setdefault('timestamp', False)
        config.setdefault('link', ['imdb_url', 'input_url'])
        config.setdefault('title', '{{title}} (from {{task}})')
        config.setdefault('template', 'rss')
        # add url as last resort
        config['link'].append('url')
        return config

    def on_task_exit(self, task, config):
        """Store finished / downloaded entries at exit"""
        if not rss2gen:
            raise plugin.PluginWarning('plugin make_rss requires PyRSS2Gen library.')
        config = self.prepare_config(config)

        # when history is disabled, remove everything from backlog on every run (a bit hackish, rarely useful)
        if not config['history']:
            logger.debug('disabling history')
            for item in task.session.query(RSSEntry).filter(RSSEntry.file == config['file']).all():
                task.session.delete(item)

        # save entries into db for RSS generation
        for entry in task.accepted:
            rss = RSSEntry()
            try:
                rss.title = entry.render(config['title'])
            except RenderError as e:
                logger.error(
                    'Error rendering jinja title for `{}` falling back to entry title: {}',
                    entry['title'],
                    e,
                )
                rss.title = entry['title']
            for field in config['link']:
                if field in entry:
                    rss.link = entry[field]
                    break

            try:
                template = get_template(config['template'], scope='task')
            except ValueError as e:
                raise plugin.PluginError('Invalid template specified: %s' % e)
            try:
                rss.description = render_from_entry(template, entry)
            except RenderError as e:
                logger.error(
                    'Error while rendering entry {}, falling back to plain title: {}', entry, e
                )
                rss.description = entry['title'] + ' - (Render Error)'
            rss.file = config['file']
            if 'rss_pubdate' in entry:
                rss.published = entry['rss_pubdate']

            rss.enc_length = entry['size'] if 'size' in entry else None
            rss.enc_type = entry['type'] if 'type' in entry else None

            # TODO: check if this exists and suggest disabling history if it does since it shouldn't happen normally ...
            logger.debug('Saving {} into rss database', entry['title'])
            task.session.add(rss)

        if not rss2gen:
            return
        # don't generate rss when learning
        if task.options.learn:
            return

        db_items = (
            task.session.query(RSSEntry)
            .filter(RSSEntry.file == config['file'])
            .order_by(RSSEntry.published.desc())
            .all()
        )

        # make items
        rss_items = []
        for db_item in db_items:
            add = True
            if config['items'] != -1:
                if len(rss_items) > config['items']:
                    add = False
            if config['days'] != -1:
                if (
                    datetime.datetime.today() - datetime.timedelta(days=config['days'])
                    > db_item.published
                ):
                    add = False
            if add:
                # add into generated feed
                hasher = hashlib.sha1()
                hasher.update(db_item.title.encode('utf8'))
                hasher.update(db_item.description.encode('utf8'))
                hasher.update(db_item.link.encode('utf8'))
                guid = base64.urlsafe_b64encode(hasher.digest()).decode('ascii')
                guid = PyRSS2Gen.Guid(guid, isPermaLink=False)

                gen = {
                    'title': db_item.title,
                    'description': db_item.description,
                    'link': db_item.link,
                    'pubDate': db_item.published,
                    'guid': guid,
                }
                if db_item.enc_length is not None and db_item.enc_type is not None:
                    gen['enclosure'] = PyRSS2Gen.Enclosure(
                        db_item.link, db_item.enc_length, db_item.enc_type
                    )
                logger.trace('Adding {} into rss {}', gen['title'], config['file'])
                rss_items.append(PyRSS2Gen.RSSItem(**gen))
            else:
                # no longer needed
                task.session.delete(db_item)

        # make rss
        rss = PyRSS2Gen.RSS2(
            title=config.get('rsstitle', 'FlexGet'),
            link=config.get('rsslink', 'http://flexget.com'),
            description=config.get('rssdesc', 'FlexGet generated RSS feed'),
            lastBuildDate=datetime.datetime.utcnow() if config['timestamp'] else None,
            items=rss_items,
        )

        # don't run with --test
        if task.options.test:
            logger.info('Would write rss file with {} entries.', len(rss_items))
            return

        # write rss
        fn = os.path.expanduser(config['file'])
        with open(fn, 'wb') as file:
            try:
                logger.verbose('Writing output rss to {}', fn)
                rss.write_xml(file, encoding=config['encoding'])
            except LookupError:
                logger.critical('Unknown encoding {}', config['encoding'])
                return
            except OSError:
                # TODO: plugins cannot raise PluginWarnings in terminate event ..
                logger.critical('Unable to write {}', fn)
                return


@event('plugin.register')
def register_plugin():
    plugin.register(OutputRSS, 'make_rss', api_ver=2)



from flask import jsonify, request, Response
from jsonschema import RefResolutionError
from sqlalchemy.orm import Session

from flexget.api import APIResource, api
from flexget.api.app import NotFoundError
from flexget.config_schema import resolve_ref, schema_paths

schema_api = api.namespace('schema', description='Config and plugin schemas')

schema_api_list = api.schema_model(
    'schema.list',
    {'type': 'object', 'properties': {'schemas': {'type': 'array', 'items': {'type': 'object'}}}},
)


def rewrite_ref(identifier: str, base_url: str) -> str:
    """
    The refs in the schemas are arbitrary identifiers, and cannot be used as-is as real network locations.
    This rewrites any of those arbitrary refs to be real urls servable by this endpoint.
    """
    if not base_url.endswith('/'):
        base_url += '/'
    if identifier.startswith('/schema/'):
        return base_url + identifier[1:]
    return identifier


def rewrite_refs(schema, base_url: str):
    """Make sure any $refs in the schema point properly back to this endpoint."""
    if isinstance(schema, dict):
        if '$ref' in schema:
            return {'$ref': rewrite_ref(schema['$ref'], base_url)}
        return {k: rewrite_refs(v, base_url) for k, v in schema.items()}
    if isinstance(schema, list):
        return [rewrite_refs(v, base_url) for v in schema]
    return schema


@schema_api.route('/')
class SchemaAllAPI(APIResource):
    @api.response(200, model=schema_api_list)
    def get(self, session: Session = None) -> Response:
        """ List all schema definitions """
        schemas = []
        for path in schema_paths:
            schema = rewrite_refs(resolve_ref(path), request.url_root)
            schema['id'] = rewrite_ref(path, request.url_root)
            schemas.append(schema)
        return jsonify({'schemas': schemas})


@schema_api.route('/<path:path>/')
@api.doc(params={'path': 'Path of schema'})
@api.response(NotFoundError)
class SchemaAPI(APIResource):
    @api.response(200, model=schema_api_list)
    def get(self, path: str, session: Session = None) -> Response:
        """ Get schema definition """
        try:
            schema = resolve_ref(request.full_path)
        except RefResolutionError:
            raise NotFoundError('invalid schema path')
        schema['id'] = request.url
        return jsonify(rewrite_refs(schema, request.url_root))



import xml.etree.ElementTree as ET

from loguru import logger
from requests.exceptions import RequestException

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import PluginWarning
from flexget.utils.requests import Session as RequestSession
from flexget.utils.requests import TimedLimiter

plugin_name = 'prowl'
logger = logger.bind(name=plugin_name)

PROWL_URL = 'https://api.prowlapp.com/publicapi/add'

requests = RequestSession(max_retries=3)
requests.add_domain_limiter(TimedLimiter('prowlapp.com', '5 seconds'))


class ProwlNotifier:
    """
    Send prowl notifications

    Example::

      notify:
        entries:
          via:
            - prowl:
                api_key: xxxxxxx
                [application: application name, default FlexGet]
                [event: event title, default New Release]
                [priority: -2 - 2 (2 = highest), default 0]
                [description: notification to send]

    """

    schema = {
        'type': 'object',
        'properties': {
            'api_key': one_or_more({'type': 'string'}),
            'application': {'type': 'string', 'default': 'FlexGet'},
            'priority': {'type': 'integer', 'minimum': -2, 'maximum': 2},
            'url': {'type': 'string'},
            'provider_key': {'type': 'string'},
        },
        'required': ['api_key'],
        'additionalProperties': False,
    }

    def notify(self, title, message, config):
        """
        Send a Prowl notification
        """
        notification = {
            'application': config.get('application'),
            'event': title,
            'description': message,
            'url': config.get('url'),
            'priority': config.get('priority'),
            'providerkey': config.get('provider_key'),
        }

        if isinstance(config['api_key'], list):
            config['api_key'] = [config['api_key']]
        notification['apikey'] = config['api_key']

        try:
            response = requests.post(PROWL_URL, data=notification)
        except RequestException as e:
            raise PluginWarning(repr(e))

        request_status = ET.fromstring(response.content)
        error = request_status.find('error')
        if error is not None:
            raise PluginWarning(error.text)
        else:
            success = request_status.find('success').attrib
            logger.debug(
                'prowl notification sent. Notifications remaining until next reset: {}. '
                'Next reset will occur in {} minutes',
                success['remaining'],
                success['resetdate'],
            )


@event('plugin.register')
def register_plugin():
    plugin.register(ProwlNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



from flexget.api.app import base_message
from flexget.api.core.plugins import ObjectsContainer as OC
from flexget.utils import json


class TestPluginsAPI:
    config = 'tasks: {}'

    def test_plugins_api(self, api_client, schema_match):
        rsp = api_client.get('/plugins/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list_reply, data)
        assert not errors

        rsp = api_client.get('/plugins/?include_schema=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list_reply, data)
        assert not errors

        rsp = api_client.get('/plugins/?interface=search')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list_reply, data)
        assert not errors

        rsp = api_client.get('/plugins/?interface=fgfg')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list_reply, data)
        assert not errors
        assert data == []

        rsp = api_client.get('/plugins/?phase=fgfg')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/plugins/?phase=input')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list_reply, data)
        assert not errors

        rsp = api_client.get('/plugins/bla/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/plugins/seen/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_object, data)
        assert not errors

        rsp = api_client.get('/plugins/seen/?include_schema=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_object, data)
        assert not errors



import re

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils.template import evaluate_expression

logger = logger.bind(name='sort_by')

RE_ARTICLES = r'^(the|a|an)\s'


class PluginSortBy:
    """
    Sort task entries based on a field

    Example::

      sort_by: title

    More complex::

      sort_by:
        field: imdb_score
        reverse: yes

    Reverse the order of the entries, without sorting on a field::

      sort_by:
        reverse: yes
    """

    schema = one_or_more(
        {
            'oneOf': [
                {'type': 'string'},
                {
                    'type': 'object',
                    'properties': {
                        'field': {'type': 'string'},
                        'reverse': {'type': 'boolean'},
                        'ignore_articles': {
                            'oneOf': [{'type': 'boolean'}, {'type': 'string', 'format': 'regex'}]
                        },
                    },
                    'additionalProperties': False,
                },
            ]
        }
    )

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_filter(self, task, config):
        if not isinstance(config, list):
            config = [config]
        # First item in the config list should be primary sort key, so iterate in reverse
        for item in reversed(config):
            if isinstance(item, str):
                field = item
                reverse = False
                ignore_articles = False
            else:
                field = item.get('field', None)
                reverse = item.get('reverse', False)
                ignore_articles = item.get('ignore_articles', False)

            logger.debug('sorting entries by: {}', item)

            if not field:
                if reverse:
                    task.all_entries.reverse()
                continue

            re_articles = RE_ARTICLES if ignore_articles is True else ignore_articles

            def sort_key(entry):
                val = evaluate_expression(field, entry)
                if isinstance(val, str) and re_articles:
                    val = re.sub(re_articles, '', val, flags=re.IGNORECASE)
                # Sort None values last no matter whether reversed or not
                return (val is not None) == reverse, val

            task.all_entries.sort(key=sort_key, reverse=reverse)


@event('plugin.register')
def register_plugin():
    plugin.register(PluginSortBy, 'sort_by', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='anirena')


class UrlRewriteAniRena:
    """AniRena urlrewriter."""

    def url_rewritable(self, task, entry):
        return entry['url'].startswith('http://www.anirena.com/viewtracker.php?action=details&id=')

    def url_rewrite(self, task, entry):
        entry['url'] = entry['url'].replace('details', 'download')


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteAniRena, 'anirena', interfaces=['urlrewriter'], api_ver=2)



SCHEMA_SERVER = {
    'oneOf': [
        {
            'type': 'object',
            'properties': {
                'host': {'type': 'string', "default": "http://localhost:8096"},
                "username": {'type': 'string'},
                "password": {'type': 'string'},
                "return_host": {'type': 'string', 'enum': ['lan', 'wan']},
            },
            'required': ['username', "password"],
            'additionalProperties': False,
        },
        {
            'type': 'object',
            'properties': {
                'host': {'type': 'string', "default": "http://localhost:8096"},
                'username': {'type': 'string'},
                "apikey": {'type': 'string'},
                "return_host": {'type': 'string', 'enum': ['lan', 'wan']},
            },
            'required': ['username', 'apikey'],
            'additionalProperties': False,
        },
    ]
}


SCHEMA_SERVER_TAG = {"server": {**SCHEMA_SERVER}}


SORT_FIELDS = [
    'comunity_rating',
    'critic_rating',
    'date_created',
    'date_played',
    'play_count',
    'premiere_date',
    'production_year',
    'sort_name',
    'random',
    'revenue',
    'runtime',
]


field_map = {
    'media': {
        'emby_id': 'id',
        'emby_name': 'name',
        'emby_fullname': 'fullname',
        'emby_type': 'mtype',
        'emby_created_date': 'created_date',
        'emby_path': 'path',
        'emby_filename': 'filename',
        'emby_file_extension': 'file_extension',
        'emby_watched': 'watched',
        'emby_favorite': 'favorite',
        'emby_playcount': 'playcount',
        'emby_media_sources_raw': 'media_sources_raw',
        'emby_format_3d': 'format_3d',
        'emby_audio': 'audio',
        'emby_quality': 'quality',
        'emby_subtitles': 'subtitles',
        'imdb_url': 'imdb_url',
        'emby_download_url': 'download_url',
        'emby_library_name': 'library_name',
        'emby_page': 'page',
    },
    'series': {
        'emby_serie_year': 'serie_year',
        'emby_serie_aired_date': 'serie_aired_date',
        'emby_serie_id': 'serie_id',
        'emby_serie_name': 'serie_name',
        'emby_serie_photo': 'serie_photo',
        'emby_serie_imdb_id': 'serie_imdb_id',
        'emby_serie_tvdb_id': 'serie_tvdb_id',
        'emby_serie_overview': 'serie_overview',
        'emby_serie_page': 'serie_page',
        'imdb_id': 'serie_imdb_id',
        'tvdb_id': 'serie_tvdb_id',
    },
    'season': {
        'emby_season': 'season',
        'emby_season_name': 'season_name',
        'emby_season_id': 'season_id',
        'emby_season_page': 'season_page',
        'emby_season_photo': 'season_photo',
        'emby_season_imdb_id': 'season_imdb_id',
        'emby_season_tmdb_id': 'season_tmdb_id',
        'emby_season_tvdb_id': 'season_tvdb_id',
    },
    'episode': {
        'emby_episode': 'episode',
        'emby_ep_name': 'ep_name',
        'emby_ep_page': 'ep_page',
        'emby_ep_id': 'ep_id',
        'emby_ep_photo': 'ep_photo',
        'emby_ep_imdb_id': 'ep_imdb_id',
        'emby_ep_tmdb_id': 'ep_tmdb_id',
        'emby_ep_tvdb_id': 'ep_tvdb_id',
        'emby_ep_aired_date': 'ep_aired_date',
        'emby_ep_overview': 'ep_overview',
    },
    'movie': {
        'movie_name': 'movie_name',
        'movie_year': 'movie_year',
        'emby_movie_name': 'movie_name',
        'emby_movie_id': 'movie_id',
        'emby_movie_imdb_id': 'movie_imdb_id',
        'emby_movie_tmdb_id': 'movie_tmdb_id',
        'emby_movie_year': 'movie_year',
        'emby_movie_aired_date': 'movie_aired_date',
        'emby_movie_photo': 'movie_photo',
        'emby_movie_page': 'movie_page',
        'emby_movie_overview': 'movie_overview',
        'imdb_id': 'movie_imdb_id',
        'tmdb_id': 'movie_tmdb_id',
    },
}


def get_field_map(**kwargs):
    lazy_fields = field_map['media']

    media_type = kwargs.get('media_type')
    if not media_type:
        media_type = kwargs.get('emby_type', '')

    media_type = media_type.lower()

    if media_type in field_map:
        lazy_fields = {**lazy_fields, **field_map[media_type]}
    else:
        lazy_fields = {**lazy_fields, **field_map['movie']}

    if media_type == 'episode':
        lazy_fields = {**lazy_fields, **field_map['series'], **field_map['season']}
    elif media_type == 'season':
        lazy_fields = {**lazy_fields, **field_map['series']}

    return lazy_fields



from loguru import logger

from flexget.components.emby.api_emby import EmbyAuth, EmbyApi
from flexget.config_schema import one_or_more
from flexget import plugin
from flexget.event import event
from flexget.components.emby.emby_util import SCHEMA_SERVER_TAG, SORT_FIELDS

logger = logger.bind(name='from_emby')


class EmbyInput:
    """
    Returns Emby Inputs

    Example:
        from_emby:
            server:
                host: http://localhost:8096
                username: <username>
                apikey: <apikey>
                return_host: wan
            list: TV
            types: series
    """

    auth = None

    schema = {
        'type': 'object',
        'properties': {
            **SCHEMA_SERVER_TAG,
            'list': {'type': 'string'},
            'types': one_or_more(
                {'type': 'string', 'enum': ['movie', 'series', 'season', 'episode']}
            ),
            'watched': {'type': 'boolean'},
            'favorite': {'type': 'boolean'},
            'sort': {
                'oneOf': [
                    {
                        'type': 'string',
                        'enum': SORT_FIELDS,
                    },
                    {
                        'type': 'object',
                        'properties': {
                            'field': {
                                'type': 'string',
                                'enum': SORT_FIELDS,
                            },
                            'order': {'type': 'string', 'enum': ['ascending', 'descending']},
                        },
                        'required': ['field', 'order'],
                    },
                ]
            },
        },
        'required': ['server', 'list'],
        'additionalProperties': False,
    }

    def login(self, config):
        if self.auth and self.auth.logged:
            return

        if not isinstance(config, dict):
            config = {}

        self.auth = EmbyAuth(**config)
        self.auth.login(True)

    def on_task_start(self, task, config):
        self.login(config)

    @plugin.internet(logger)
    def on_task_input(self, task, config):
        self.login(config)

        s_lists = EmbyApi.search_list(**config)

        for s_list in s_lists:
            entry = s_list.to_entry()
            yield entry


@event('plugin.register')
def register_plugin():
    plugin.register(EmbyInput, 'from_emby', api_ver=2)



import logging

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.plugin import DependencyError, PluginWarning

plugin_name = 'xmpp'

logger = logger.bind(name=plugin_name)


class XMPPNotifier:
    """
    Sends messages via XMPP. The sleekxmpp library is required to be installed.
    Install it with: `pip install sleekxmpp`

    All fields are required.

    Example::

      notify:
        entries:
          via:
            - xmpp:
                sender: sender's JID
                password: sender's password
                recipients: recipient's JID or list of JIDs
    """

    schema = {
        'type': 'object',
        'properties': {
            'sender': {'type': 'string'},
            'password': {'type': 'string'},
            'recipients': one_or_more({'type': 'string'}),
        },
        'required': ['sender', 'password', 'recipients'],
        'additionalProperties': False,
    }

    __version__ = '1.0'

    def notify(self, title, message, config):
        try:
            import sleekxmpp  # noqa
        except ImportError as e:
            logger.debug('Error importing SleekXMPP: {}', e)
            raise DependencyError(
                plugin_name, 'sleekxmpp', 'SleekXMPP module required. ImportError: %s' % e
            )
        try:
            import dns  # noqa
        except ImportError:
            try:
                import dnspython  # noqa
            except ImportError as e:
                logger.debug('Error importing dnspython: {}', e)
                raise DependencyError(
                    plugin_name, 'dnspython', 'dnspython module required. ImportError: %s' % e
                )

        class SendMsgBot(sleekxmpp.ClientXMPP):
            def __init__(self, jid, password, recipients, message):
                sleekxmpp.ClientXMPP.__init__(self, jid, password)
                self.recipients = recipients
                self.msg = message
                self.add_event_handler("session_start", self.start, threaded=True)
                self.register_plugin('xep_0030')  # Service Discovery
                self.register_plugin('xep_0199')  # XMPP Ping

            def start(self, xmpp_event):
                for recipient in self.recipients:
                    self.send_presence(pto=recipient)
                    self.send_message(mto=recipient, mbody=self.msg, mtype='chat')
                self.disconnect(wait=True)

        message = '%s\n%s' % (title, message)
        logger.debug('Sending XMPP notification about: {}', message)
        logging.getLogger('sleekxmpp').setLevel(logging.CRITICAL)

        if not isinstance(config['recipients'], list):
            config['recipients'] = [config['recipients']]

        xmpp = SendMsgBot(config['sender'], config['password'], config['recipients'], message)
        if xmpp.connect():
            xmpp.process(block=True)


@event('plugin.register')
def register_plugin():
    plugin.register(XMPPNotifier, plugin_name, api_ver=2, interfaces=['notifiers'])



import json
import os
import re
from collections import deque
from contextlib import suppress
from functools import partial, wraps
from typing import Callable, List, Dict, Union, Mapping, Tuple, TYPE_CHECKING

from jsonschema.exceptions import ValidationError as SchemaValidationError
from flask import Flask, Response, jsonify, make_response, request, Request
from flask_compress import Compress
from flask_cors import CORS
from flask_restx import Api as RestxAPI
from flask_restx import Resource, Model
from flask_restx.reqparse import RequestParser
from jsonschema import RefResolutionError
from loguru import logger
from sqlalchemy.orm import Session
from werkzeug.http import generate_etag

from flexget import manager
from flexget.config_schema import format_checker, process_config
from flexget.utils.database import with_session
from flexget.webserver import User

from . import __path__

__version__ = '1.8.0'

logger = logger.bind(name='api')

if TYPE_CHECKING:
    from typing import TypedDict

    _TypeDict = TypedDict('_TypeDict', {'type': str})

    class MessageDict(_TypeDict):
        properties: Dict[str, _TypeDict]
        required: List[str]

    PaginationHeaders = TypedDict(
        'PaginationHeaders', {'Link': str, 'Total-Count': int, 'Count': int}
    )


class APIClient:
    """
    This is an client which can be used as a more pythonic interface to the rest api.

    It skips http, and is only usable from within the running flexget process.
    """

    def __init__(self) -> None:
        self.app = api_app.test_client()

    def __getattr__(self, item: str) -> 'APIEndpoint':
        return APIEndpoint('/api/' + item, self.get_endpoint)

    def get_endpoint(self, url: str, data=None, method: str = None):
        if method is None:
            method = 'POST' if data is not None else 'GET'
        auth_header = dict(Authorization='Token %s' % api_key())
        response = self.app.open(
            url, data=data, follow_redirects=True, method=method, headers=auth_header
        )
        result = json.loads(response.get_data(as_text=True))
        # TODO: Proper exceptions
        if 200 > response.status_code >= 300:
            raise Exception(result['error'])
        return result


class APIEndpoint:
    def __init__(self, endpoint: str, caller: Callable) -> None:
        self.endpoint = endpoint
        self.caller = caller

    def __getattr__(self, item):
        return self.__class__(self.endpoint + '/' + item, self.caller)

    __getitem__ = __getattr__

    def __call__(self, data=None, method: str = None):
        return self.caller(self.endpoint, data=data, method=method)


def api_version(f: Callable[..., Response]):
    """ Add the 'API-Version' header to all responses """

    @wraps(f)
    def wrapped(*args, **kwargs):
        rv = f(*args, **kwargs)
        rv.headers['API-Version'] = __version__
        return rv

    return wrapped


class APIResource(Resource):
    """All api resources should subclass this class."""

    method_decorators = [with_session, api_version]

    def __init__(self, api: RestxAPI, *args, **kwargs) -> None:
        self.manager = manager.manager
        super().__init__(api, *args, **kwargs)


class API(RestxAPI):
    """
    Extends a flask restx :class:`flask_restx.Api` with:
      - methods to make using json schemas easier
      - methods to auto document and handle :class:`ApiError` responses
    """

    def validate(
        self,
        model: Model,
        schema_override: Dict[str, List[Dict[str, str]]] = None,
        description=None,
    ):
        """
        When a method is decorated with this, json data submitted to the endpoint will be validated with the given
        `model`. This also auto-documents the expected model, as well as the possible :class:`ValidationError` response.
        """

        def decorator(func):
            @api.expect((model, description))
            @api.response(ValidationError)
            @wraps(func)
            def wrapper(*args, **kwargs):
                payload = request.json
                try:
                    schema = schema_override if schema_override else model.__schema__
                    errors = process_config(config=payload, schema=schema, set_defaults=False)

                    if errors:
                        raise ValidationError(errors)
                except RefResolutionError as e:
                    raise APIError(str(e))
                return func(*args, **kwargs)

            return wrapper

        return decorator

    def response(self, code_or_apierror, description: str = 'Success', model=None, **kwargs):
        """
        Extends :meth:`flask_restx.Api.response` to allow passing an :class:`ApiError` class instead of
        response code. If an `ApiError` is used, the response code, and expected response model, is automatically
        documented.
        """
        with suppress(TypeError):  # If first argument isn't a class this happens
            if issubclass(code_or_apierror, APIError):
                description = code_or_apierror.description or description
                return self.doc(
                    responses={
                        code_or_apierror.status_code: (
                            description,
                            code_or_apierror.response_model,
                        )
                    },
                    **kwargs,
                )
        return self.doc(responses={code_or_apierror: (description, model)}, **kwargs)

    def pagination_parser(
        self,
        parser: RequestParser = None,
        sort_choices: List[str] = None,
        default: str = None,
        add_sort: bool = None,
    ) -> RequestParser:
        """
        Return a standardized pagination parser, to be used for any endpoint that has pagination.

        :param RequestParser parser: Can extend a given parser or create a new one
        :param tuple sort_choices: A tuple of strings, to be used as server side attribute searches
        :param str default: The default sort string, used `sort_choices[0]` if not given
        :param bool add_sort: Add sort order choices without adding specific sort choices

        :return: An api.parser() instance with pagination and sorting arguments.
        """
        pagination = parser.copy() if parser else self.parser()
        pagination.add_argument('page', type=int, default=1, help='Page number')
        pagination.add_argument('per_page', type=int, default=50, help='Results per page')
        if sort_choices or add_sort:
            pagination.add_argument(
                'order', choices=('desc', 'asc'), default='desc', help='Sorting order'
            )
        if sort_choices:
            pagination.add_argument(
                'sort_by',
                choices=sort_choices,
                default=default or sort_choices[0],
                help='Sort by attribute',
            )

        return pagination


api_app = Flask(__name__, template_folder=os.path.join(__path__[0], 'templates'))
api_app.config['REMEMBER_COOKIE_NAME'] = 'flexget.token'
api_app.config['DEBUG'] = True
api_app.config['ERROR_404_HELP'] = False
api_app.url_map.strict_slashes = False

CORS(api_app, expose_headers='Link, Total-Count, Count, ETag')
Compress(api_app)

api = API(
    api_app,
    title='Flexget API v{}'.format(__version__),
    version=__version__,
    description='View and manage flexget core operations and plugins. Open each endpoint view for usage information.'
    ' Navigate to http://flexget.com/API for more details.',
    format_checker=format_checker,
)

base_message: "MessageDict" = {
    'type': 'object',
    'properties': {
        'status_code': {'type': 'integer'},
        'message': {'type': 'string'},
        'status': {'type': 'string'},
    },
    'required': ['status_code', 'message', 'status'],
}

base_message_schema = api.schema_model('base_message', base_message)


class APIError(Exception):
    description = 'Server error'
    status_code = 500
    status = 'Error'
    response_model = base_message_schema

    def __init__(self, message: str = None, payload=None) -> None:
        self.message = message
        self.payload = payload

    def to_dict(self):
        rv = self.payload or {}
        rv.update(status_code=self.status_code, message=self.message, status=self.status)
        return rv

    @classmethod
    def schema(cls):
        return cls.response_model.__schema__


class NotFoundError(APIError):
    status_code = 404
    description = 'Not found'


class Unauthorized(APIError):
    status_code = 401
    description = 'Unauthorized'


class BadRequest(APIError):
    status_code = 400
    description = 'Bad request'


class Conflict(APIError):
    status_code = 409
    description = 'Conflict'


class PreconditionFailed(APIError):
    status_code = 412
    description = 'Precondition failed'


class NotModified(APIError):
    status_code = 304
    description = 'not modified'


class ValidationError(APIError):
    status_code = 422
    description = 'Validation error'

    response_model = api.schema_model(
        'validation_error',
        {
            'type': 'object',
            'properties': {
                'validation_errors': {
                    'type': 'array',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'message': {
                                'type': 'string',
                                'description': 'A human readable message explaining the error.',
                            },
                            'validator': {
                                'type': 'string',
                                'description': 'The name of the failed validator.',
                            },
                            'validator_value': {
                                'type': 'string',
                                'description': 'The value for the failed validator in the schema.',
                            },
                            'path': {'type': 'string'},
                            'schema_path': {'type': 'string'},
                        },
                    },
                }
            },
            'required': ['validation_errors'],
        },
    )

    verror_attrs = (
        'message',
        'cause',
        'validator',
        'validator_value',
        'path',
        'schema_path',
        'parent',
    )

    def __init__(
        self, validation_errors: List[SchemaValidationError], message: str = 'validation error'
    ) -> None:
        payload = {
            'validation_errors': [self._verror_to_dict(error) for error in validation_errors]
        }
        super().__init__(message, payload=payload)

    def _verror_to_dict(self, error: SchemaValidationError) -> Mapping[str, Union[str, list]]:
        error_dict: Dict[str, Union[str, list]] = {}
        for attr in self.verror_attrs:
            if isinstance(getattr(error, attr), deque):
                error_dict[attr] = list(getattr(error, attr))
            else:
                error_dict[attr] = str(getattr(error, attr))
        return error_dict


empty_response = api.schema_model('empty', {'type': 'object'})


def success_response(message: str, status_code: int = 200, status: str = 'success') -> Response:
    rsp_dict = {'message': message, 'status_code': status_code, 'status': status}
    rsp = jsonify(rsp_dict)
    rsp.status_code = status_code
    return rsp


@api.errorhandler(APIError)
@api.errorhandler(NotFoundError)
@api.errorhandler(ValidationError)
@api.errorhandler(BadRequest)
@api.errorhandler(Unauthorized)
@api.errorhandler(Conflict)
@api.errorhandler(NotModified)
@api.errorhandler(PreconditionFailed)
def api_errors(error: APIError) -> Tuple[dict, int]:
    return error.to_dict(), error.status_code


@with_session
def api_key(session: Session = None) -> str:
    logger.debug('fetching token for internal lookup')
    return session.query(User).first().token  # type: ignore


def etag(method: Callable = None, cache_age: int = 0):
    """
    A decorator that add an ETag header to the response and checks for the "If-Match" and "If-Not-Match" headers to
     return an appropriate response.

    :param method: A GET or HEAD flask method to wrap
    :param cache_age: max-age cache age for the content
    :return: The method's response with the ETag and Cache-Control headers, raises a 412 error or returns a 304 response
    """

    # If called without method, we've been called with optional arguments.
    # We return a decorator with the optional arguments filled in.
    # Next time round we'll be decorating method.
    if method is None:
        return partial(etag, cache_age=cache_age)

    @wraps(method)
    def wrapped(*args, **kwargs):
        # Identify if this is a GET or HEAD in order to proceed
        assert request.method in ['HEAD', 'GET'], '@etag is only supported for GET requests'
        rv = method(*args, **kwargs)
        rv = make_response(rv)

        # Some headers can change without data change for specific page
        content_headers = (
            rv.headers.get('link', '')
            + rv.headers.get('count', '')
            + rv.headers.get('total-count', '')
        )
        data = (rv.get_data().decode() + content_headers).encode()
        etag = generate_etag(data)
        rv.headers['Cache-Control'] = 'max-age=%s' % cache_age
        rv.headers['ETag'] = etag
        if_match = request.headers.get('If-Match')
        if_none_match = request.headers.get('If-None-Match')

        if if_match:
            etag_list = [tag.strip() for tag in if_match.split(',')]
            if etag not in etag_list and '*' not in etag_list:
                raise PreconditionFailed('etag does not match')
        elif if_none_match:
            etag_list = [tag.strip() for tag in if_none_match.split(',')]
            if etag in etag_list or '*' in etag_list:
                raise NotModified

        return rv

    return wrapped


def pagination_headers(
    total_pages: int, total_items: int, page_count: int, request: Request
) -> 'PaginationHeaders':
    """
    Creates the `Link`. 'Count' and  'Total-Count' headers, to be used for pagination traversing

    :param total_pages: Total number of pages
    :param total_items: Total number of items in all the pages
    :param page_count: Item count for page (may differ from page size request)
    :param request: The flask request used, required to build other reoccurring vars like url and such.
    :return:
    """

    # Build constant variables from request data
    url = request.url_root + request.path.lstrip('/')
    per_page = request.args.get('per_page', 50)
    page = int(request.args.get('page', 1))

    # Build the base template
    link_with_params = f'<{url}?per_page={per_page}&'

    # Removed page and per_page from query string
    query_string = re.sub(br'per_page=\d+', b'', request.query_string)
    query_string = re.sub(br'page=\d+', b'', query_string)
    query_string = re.sub(b'&{2,}', b'&', query_string)

    # Add all original query params
    link_with_params += query_string.decode().lstrip('&')

    link_string = ''

    if page > 1:
        link_string += f'{link_with_params}&page={page-1}>; rel="prev", '
    if page < total_pages:
        link_string += f'{link_with_params}&page={page+1}>; rel="next", '
    link_string += f'{link_with_params}&page={total_pages}>; rel="last"'

    return {'Link': link_string, 'Total-Count': total_items, 'Count': page_count}



"""
Miscellaneous SQLAlchemy helpers.
"""
from typing import Any, List, Optional, Union

from loguru import logger
from sqlalchemy import ColumnDefault, Index, Sequence
from sqlalchemy.exc import NoSuchTableError, OperationalError
from sqlalchemy.orm import Session
from sqlalchemy.schema import MetaData, Table
from sqlalchemy.types import TypeEngine

logger = logger.bind(name='sql_utils')


def table_exists(name: str, session: Session) -> bool:
    """
    Use SQLAlchemy reflect to check table existences.

    :param string name: Table name to check
    :param Session session: Session to use
    :return: True if table exists, False otherwise
    :rtype: bool
    """
    try:
        table_schema(name, session)
    except NoSuchTableError:
        return False
    return True


def table_schema(name: str, session: Session) -> Table:
    """
    :returns: Table schema using SQLAlchemy reflect as it currently exists in the db
    :rtype: Table
    """
    return Table(name, MetaData(bind=session.bind), autoload=True)


def table_columns(table: Union[str, Table], session: Session) -> List[str]:
    """
    :param string table: Name of table or table schema
    :param Session session: SQLAlchemy Session
    :returns: List of column names in the table or empty list
    """

    res = []
    if isinstance(table, str):
        table = table_schema(table, session)
    for column in table.columns:
        res.append(column.name)
    return res


def table_add_column(
    table: Union[Table, str],
    name: str,
    col_type: Union[TypeEngine, type],
    session: Session,
    default: Any = None,
) -> None:
    """Adds a column to a table

    .. warning:: Uses raw statements, probably needs to be changed in
                 order to work on other databases besides SQLite

    :param string table: Table to add column to (can be name or schema)
    :param string name: Name of new column to add
    :param col_type: The sqlalchemy column type to add
    :param Session session: SQLAlchemy Session to do the alteration
    :param default: Default value for the created column (optional)
    """
    if isinstance(table, str):
        table = table_schema(table, session)
    if name in table_columns(table, session):
        # If the column already exists, we don't have to do anything.
        return
    # Add the column to the table
    if not isinstance(col_type, TypeEngine):
        # If we got a type class instead of an instance of one, instantiate it
        col_type = col_type()
    type_string = session.bind.engine.dialect.type_compiler.process(col_type)
    statement = 'ALTER TABLE %s ADD %s %s' % (table.name, name, type_string)
    session.execute(statement)
    session.commit()
    # Update the table with the default value if given
    if default is not None:
        # Get the new schema with added column
        table = table_schema(table.name, session)
        if not isinstance(default, (ColumnDefault, Sequence)):
            default = ColumnDefault(default)
        default._set_parent(getattr(table.c, name))
        statement = table.update().values({name: default.execute(bind=session.bind)})
        session.execute(statement)
        session.commit()


def drop_tables(names: List[str], session: Session) -> None:
    """Takes a list of table names and drops them from the database if they exist."""
    metadata = MetaData()
    metadata.reflect(bind=session.bind)
    for table in metadata.sorted_tables:
        if table.name in names:
            table.drop()


def get_index_by_name(table: Table, name: str) -> Optional[Index]:
    """
    Find declaratively defined index from table by name

    :param table: Table object
    :param string name: Name of the index to get
    :return: Index object
    """
    for index in table.indexes:
        if index.name == name:
            return index


def create_index(table_name: str, session: Session, *column_names: str) -> None:
    """
    Creates an index on specified `columns` in `table_name`

    :param table_name: Name of table to create the index on.
    :param session: Session object which should be used
    :param column_names: The names of the columns that should belong to this index.
    """
    index_name = '_'.join(['ix', table_name] + list(column_names))
    table = table_schema(table_name, session)
    columns = [getattr(table.c, column) for column in column_names]
    try:
        Index(index_name, *columns).create(bind=session.bind)
    except OperationalError:
        logger.opt(exception=True).debug('Error creating index.')


class ContextSession(Session):
    """:class:`sqlalchemy.orm.Session` which can be used as context manager"""

    def __enter__(self) -> 'ContextSession':
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.commit()
            else:
                self.rollback()
        finally:
            self.close()



from datetime import datetime

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='est_movies')


class EstimatesReleasedMovies:
    @plugin.priority(0)
    def estimate(self, entry):
        if 'tmdb_released' in entry:
            logger.verbose('Querying release estimation for {}', entry['title'])
            return entry['tmdb_released']
        elif 'movie_year' in entry and entry['movie_year'] is not None:
            try:
                return datetime(year=entry['movie_year'], month=1, day=1)
            except ValueError:
                pass
        logger.debug(
            'Unable to check release for {}, tmdb_release and movie_year fields are not defined',
            entry['title'],
        )


@event('plugin.register')
def register_plugin():
    plugin.register(
        EstimatesReleasedMovies, 'est_released_movies', interfaces=['estimate_release'], api_ver=2
    )



import base64
import os
import re
from datetime import datetime, timedelta
from fnmatch import fnmatch
from functools import partial
from netrc import NetrcParseError, netrc
from time import sleep
from urllib.parse import urlparse

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.pathscrub import pathscrub
from flexget.utils.template import RenderError
from flexget.utils.tools import parse_timedelta

try:
    import requests.exceptions
    import transmission_rpc as transmissionrpc
    from transmission_rpc import TransmissionError
except ImportError:
    # If transmissionrpc is not found, errors will be shown later
    pass

logger = logger.bind(name='transmission')


class TransmissionBase:
    def __init__(self):
        self.client = None
        self.opener = None

    def prepare_config(self, config):
        if isinstance(config, bool):
            config = {'enabled': config}
        config.setdefault('enabled', True)
        config.setdefault('host', 'localhost')
        config.setdefault('port', 9091)
        config.setdefault('main_file_ratio', 0.9)
        if 'netrc' in config:
            netrc_path = os.path.expanduser(config['netrc'])
            try:
                config['username'], _, config['password'] = netrc(netrc_path).authenticators(
                    config['host']
                )
            except OSError as e:
                logger.error('netrc: unable to open: {}', e.filename)
            except NetrcParseError as e:
                logger.error('netrc: {}, file: {}, line: {}', e.msg, e.filename, e.lineno)
        return config

    def create_rpc_client(self, config):
        user, password = config.get('username'), config.get('password')
        urlo = urlparse(config['host'])

        if urlo.scheme == '':
            urlo = urlparse('http://' + config['host'])

        protocol = urlo.scheme if urlo.scheme else 'http'
        port = str(urlo.port) if urlo.port else config['port']
        path = urlo.path.rstrip('rpc') if urlo.path else '/transmission/'

        logger.debug('Connecting to {}://{}:{}{}', protocol, urlo.hostname, port, path)

        try:
            cli = transmissionrpc.Client(
                protocol=protocol,
                host=urlo.hostname,
                port=port,
                path=path,
                username=user,
                password=password,
            )
        except TransmissionError as e:
            if e.original and e.original.code == 401:
                raise plugin.PluginError(
                    "Username/password for transmission is incorrect. Cannot connect."
                )
            else:
                raise plugin.PluginError("Error connecting to transmission: %s" % e.message)
        except requests.exceptions.ConnectTimeout as e:
            raise plugin.PluginError("Cannot connect to transmission: Connection timed out.")
        except requests.exceptions.ConnectionError as e:
            raise plugin.PluginError("Error connecting to transmission: %s" % e.args[0].reason)
        except ValueError as e:
            raise plugin.PluginError("Error connecting to transmission")
        return cli

    def torrent_info(self, torrent, config):
        done = torrent.totalSize > 0
        vloc = None
        best = None
        for t in torrent.files().items():
            tf = t[1]
            if tf['selected']:
                if tf['size'] <= 0 or tf['completed'] < tf['size']:
                    done = False
                    break
                if not best or tf['size'] > best[1]:
                    best = (tf['name'], tf['size'])
        if (
            done
            and best
            and (100 * float(best[1]) / float(torrent.totalSize))
            >= (config['main_file_ratio'] * 100)
        ):
            vloc = ('%s/%s' % (torrent.downloadDir, best[0])).replace('/', os.sep)
        return done, vloc

    def check_seed_limits(self, torrent, session):
        seed_limit_ok = True  # will remain if no seed ratio defined
        idle_limit_ok = True  # will remain if no idle limit defined

        if torrent.seedRatioMode == 1:  # use torrent's own seed ratio limit
            seed_limit_ok = torrent.uploadRatio >= torrent.seedRatioLimit
        elif torrent.seedRatioMode == 0:  # use global rules
            if session.seedRatioLimited:
                seed_limit_ok = torrent.uploadRatio >= session.seedRatioLimit

        if torrent.seedIdleMode == 1:  # use torrent's own idle limit
            idle_limit_ok = (
                torrent.date_active + timedelta(minutes=torrent.seedIdleLimit) < datetime.now()
            )
        elif torrent.seedIdleMode == 0:  # use global rules
            if session.idle_seeding_limit_enabled:
                idle_limit_ok = (
                    torrent.date_active + timedelta(minutes=session.idle_seeding_limit)
                    < datetime.now()
                )

        return seed_limit_ok, idle_limit_ok

    def on_task_start(self, task, config):
        try:
            import transmission_rpc as transmissionrpc
            from transmission_rpc import TransmissionError  # noqa
        except:
            raise plugin.PluginError(
                'transmission-rpc module version 3.0 or higher required.', logger
            )

        # Mark rpc client for garbage collector so every task can start
        # a fresh new according its own config - fix to bug #2804
        self.client = None
        config = self.prepare_config(config)
        if config['enabled']:
            if task.options.test:
                logger.info('Trying to connect to transmission...')
                self.client = self.create_rpc_client(config)
                if self.client:
                    logger.info('Successfully connected to transmission.')
                else:
                    logger.error('It looks like there was a problem connecting to transmission.')


class PluginTransmissionInput(TransmissionBase):
    schema = {
        'anyOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'host': {'type': 'string'},
                    'port': {'type': 'integer'},
                    'netrc': {'type': 'string', 'format': 'file'},
                    'username': {'type': 'string'},
                    'password': {'type': 'string'},
                    'enabled': {'type': 'boolean'},
                    'only_complete': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        config = TransmissionBase.prepare_config(self, config)
        config.setdefault('only_complete', False)
        return config

    def on_task_input(self, task, config):
        config = self.prepare_config(config)
        if not config['enabled']:
            return

        if not self.client:
            self.client = self.create_rpc_client(config)
        entries = []

        session = self.client.get_session()

        for torrent in self.client.get_torrents():
            seed_ratio_ok, idle_limit_ok = self.check_seed_limits(torrent, session)
            if config['only_complete'] and not (
                seed_ratio_ok and idle_limit_ok and torrent.progress == 100
            ):
                continue
            entry = Entry(
                title=torrent.name,
                url='',
                torrent_info_hash=torrent.hashString,
                content_size=torrent.totalSize / (1024 * 1024),
            )
            # Location of torrent is only valid if transmission is on same machine as flexget
            if config['host'] in ('localhost', '127.0.0.1'):
                entry['location'] = torrent.torrentFile
                entry['url'] = 'file://' + torrent.torrentFile
            for attr in [
                'id',
                'comment',
                'desiredAvailable',
                'downloadDir',
                'isFinished',
                'isPrivate',
                'leftUntilDone',
                'ratio',
                'status',
                'date_active',
                'date_added',
                'date_done',
                'date_started',
                'errorString',
                'priority',
                'progress',
                'secondsDownloading',
                'secondsSeeding',
                'torrentFile',
            ]:
                try:
                    entry['transmission_' + attr] = getattr(torrent, attr)
                except Exception:
                    logger.opt(exception=True).debug(
                        'error when requesting transmissionrpc attribute {}', attr
                    )
            # Availability in percent
            entry['transmission_availability'] = (
                (torrent.desiredAvailable / torrent.leftUntilDone) if torrent.leftUntilDone else 0
            )

            entry['transmission_trackers'] = [t['announce'] for t in torrent.trackers]
            entry['transmission_seed_ratio_ok'] = seed_ratio_ok
            entry['transmission_idle_limit_ok'] = idle_limit_ok
            st_error_to_desc = {
                0: 'OK',
                1: 'tracker_warning',
                2: 'tracker_error',
                3: 'local_error',
            }
            entry['transmission_error_state'] = st_error_to_desc[torrent.error]
            # Built in done_date doesn't work when user adds an already completed file to transmission
            if torrent.progress == 100:
                entry['transmission_date_done'] = datetime.fromtimestamp(
                    max(torrent.addedDate, torrent.doneDate)
                )
            entries.append(entry)
        return entries


class PluginTransmission(TransmissionBase):
    """
    Add url from entry url to transmission

    Example::

      transmission:
        host: localhost
        port: 9091
        netrc: /home/flexget/.tmnetrc
        username: myusername
        password: mypassword
        path: the download location

    Default values for the config elements::

      transmission:
        host: localhost
        port: 9091
        enabled: yes
    """

    schema = {
        'anyOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'host': {'type': 'string'},
                    'port': {'type': 'integer'},
                    'netrc': {'type': 'string'},
                    'username': {'type': 'string'},
                    'password': {'type': 'string'},
                    'action': {
                        'type': 'string',
                        'enum': ['add', 'remove', 'purge', 'pause', 'resume', 'bypass_queue'],
                    },
                    'path': {'type': 'string'},
                    'max_up_speed': {'type': 'number'},
                    'max_down_speed': {'type': 'number'},
                    'max_connections': {'type': 'integer'},
                    'ratio': {'type': 'number'},
                    'add_paused': {'type': 'boolean'},
                    'content_filename': {'type': 'string'},
                    'main_file_only': {'type': 'boolean'},
                    'main_file_ratio': {'type': 'number'},
                    'magnetization_timeout': {'type': 'integer'},
                    'enabled': {'type': 'boolean'},
                    'include_subs': {'type': 'boolean'},
                    'bandwidth_priority': {'type': 'number'},
                    'honor_limits': {'type': 'boolean'},
                    'include_files': one_or_more({'type': 'string'}),
                    'skip_files': one_or_more({'type': 'string'}),
                    'rename_like_files': {'type': 'boolean'},
                    'queue_position': {'type': 'integer'},
                },
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        config = TransmissionBase.prepare_config(self, config)
        config.setdefault('action', 'add')
        config.setdefault('path', '')
        config.setdefault('main_file_only', False)
        config.setdefault('magnetization_timeout', 0)
        config.setdefault('include_subs', False)
        config.setdefault('rename_like_files', False)
        config.setdefault('include_files', [])
        return config

    @plugin.priority(120)
    def on_task_download(self, task, config):
        """
        Call download plugin to generate the temp files we will load
        into deluge then verify they are valid torrents
        """
        config = self.prepare_config(config)
        if not config['enabled']:
            return
        # If the download plugin is not enabled, we need to call it to get our temp .torrent files
        if 'download' not in task.config:
            download = plugin.get('download', self)
            for entry in task.accepted:
                if entry.get('transmission_id'):
                    # The torrent is already loaded in deluge, we don't need to get anything
                    continue
                if config['action'] != 'add' and entry.get('torrent_info_hash'):
                    # If we aren't adding the torrent new, all we need is info hash
                    continue
                download.get_temp_file(task, entry, handle_magnets=True, fail_html=True)

    @plugin.priority(135)
    def on_task_output(self, task, config):
        config = self.prepare_config(config)
        # don't add when learning
        if task.options.learn:
            return
        if not config['enabled']:
            return
        # Do not run if there is nothing to do
        if not task.accepted:
            return
        if self.client is None:
            self.client = self.create_rpc_client(config)
            if self.client:
                logger.debug('Successfully connected to transmission.')
            else:
                raise plugin.PluginError("Couldn't connect to transmission.")
        session_torrents = self.client.get_torrents()
        for entry in task.accepted:
            if task.options.test:
                logger.info('Would {} {} in transmission.', config['action'], entry['title'])
                continue
            # Compile user options into appropriate dict
            options = self._make_torrent_options_dict(config, entry)
            torrent_info = None
            for t in session_torrents:
                if t.hashString.lower() == entry.get(
                    'torrent_info_hash', ''
                ).lower() or t.id == entry.get('transmission_id'):
                    torrent_info = t
                    logger.debug(
                        'Found {} already loaded in transmission as {}',
                        entry['title'],
                        torrent_info.name,
                    )
                    break

            if not torrent_info:
                if config['action'] != 'add':
                    logger.warning(
                        'Cannot {} {} because it is not loaded in transmission.',
                        config['action'],
                        entry['title'],
                    )
                    continue
                downloaded = not entry['url'].startswith('magnet:')

                # Check that file is downloaded
                if downloaded and 'file' not in entry:
                    entry.fail('`file` field missing?')
                    continue

                # Verify the temp file exists
                if downloaded and not os.path.exists(entry['file']):
                    tmp_path = os.path.join(task.manager.config_base, 'temp')
                    logger.debug('entry: {}', entry)
                    logger.debug('temp: {}', ', '.join(os.listdir(tmp_path)))
                    entry.fail("Downloaded temp file '%s' doesn't exist!?" % entry['file'])
                    continue

                try:
                    if downloaded:
                        with open(entry['file'], 'rb') as f:
                            filedump = base64.b64encode(f.read()).decode('utf-8')
                        torrent_info = self.client.add_torrent(filedump, 30, **options['add'])
                    else:
                        if options['post'].get('magnetization_timeout', 0) > 0:
                            options['add']['paused'] = False
                        torrent_info = self.client.add_torrent(
                            entry['url'], timeout=30, **options['add']
                        )
                except TransmissionError as e:
                    logger.opt(exception=True).debug('TransmissionError')
                    logger.debug('Failed options dict: {}', options['add'])
                    msg = 'Error adding {} to transmission. TransmissionError: {}'.format(
                        entry['title'], e.message or 'N/A'
                    )
                    logger.error(msg)
                    entry.fail(msg)
                    continue
                logger.info('"{}" torrent added to transmission', entry['title'])
                # The info returned by the add call is incomplete, refresh it
                torrent_info = self.client.get_torrent(torrent_info.id)
            else:
                # Torrent already loaded in transmission
                if options['add'].get('download_dir'):
                    logger.verbose(
                        'Moving {} to "{}"', torrent_info.name, options['add']['download_dir']
                    )
                    # Move data even if current reported torrent location matches new location
                    # as transmission may fail to automatically move completed file to final
                    # location but continue reporting final location instead of real location.
                    # In such case this will kick transmission to really move data.
                    # If data is already located at new location then transmission just ignore
                    # this command.
                    self.client.move_torrent_data(
                        torrent_info.id, options['add']['download_dir'], 120
                    )

            try:
                total_size = torrent_info.totalSize
                main_id = None
                find_main_file = (
                    options['post'].get('main_file_only') or 'content_filename' in options['post']
                )
                skip_files = options['post'].get('skip_files')
                # We need to index the files if any of the following are defined
                if find_main_file or skip_files:
                    file_list = self.client.get_files(torrent_info.id)[torrent_info.id]

                    if options['post'].get('magnetization_timeout', 0) > 0 and not file_list:
                        logger.debug(
                            'Waiting {} seconds for "{}" to magnetize',
                            options['post']['magnetization_timeout'],
                            entry['title'],
                        )
                        for _ in range(options['post']['magnetization_timeout']):
                            sleep(1)
                            file_list = self.client.get_files(torrent_info.id)[torrent_info.id]
                            if file_list:
                                total_size = self.client.get_torrent(
                                    torrent_info.id, ['id', 'totalSize']
                                ).totalSize
                                break
                        else:
                            logger.warning(
                                '"{}" did not magnetize before the timeout elapsed, file list unavailable for processing.',
                                entry['title'],
                            )

                    # Find files based on config
                    dl_list = []
                    skip_list = []
                    main_list = []
                    ext_list = ['*.srt', '*.sub', '*.idx', '*.ssa', '*.ass']

                    main_ratio = config['main_file_ratio']
                    if 'main_file_ratio' in options['post']:
                        main_ratio = options['post']['main_file_ratio']

                    for file_id, file in enumerate(file_list):
                        # No need to set main_id if we're not going to need it
                        if find_main_file and file.size > total_size * main_ratio:
                            main_id = file_id

                        if 'include_files' in options['post']:
                            if any(
                                fnmatch(file.name, mask)
                                for mask in options['post']['include_files']
                            ):
                                dl_list.append(file_id)
                            elif options['post'].get('include_subs') and any(
                                fnmatch(file.name, mask) for mask in ext_list
                            ):
                                dl_list.append(file_id)

                        if skip_files:
                            if any(fnmatch(file.name, mask) for mask in skip_files):
                                skip_list.append(file_id)

                    if main_id is not None:
                        # Look for files matching main ID title but with a different extension
                        if options['post'].get('rename_like_files'):
                            for file_id, file in enumerate(file_list):
                                # if this filename matches main filename we want to rename it as well
                                fs = os.path.splitext(file.name)
                                if fs[0] == os.path.splitext(file_list[main_id].name)[0]:
                                    main_list.append(file_id)
                        else:
                            main_list = [main_id]

                        if main_id not in dl_list:
                            dl_list.append(main_id)
                    elif find_main_file:
                        logger.warning(
                            'No files in "{}" are > {:.0f}% of content size, no files renamed.',
                            entry['title'],
                            main_ratio * 100,
                        )

                    # If we have a main file and want to rename it and associated files
                    if 'content_filename' in options['post'] and main_id is not None:
                        if 'download_dir' not in options['add']:
                            download_dir = self.client.get_session().download_dir
                        else:
                            download_dir = options['add']['download_dir']

                        # Get new filename without ext
                        file_ext = os.path.splitext(file_list[main_id].name)[1]
                        file_path = os.path.dirname(
                            os.path.join(download_dir, file_list[main_id].name)
                        )
                        filename = options['post']['content_filename']
                        if config['host'] == 'localhost' or config['host'] == '127.0.0.1':
                            counter = 1
                            while os.path.exists(os.path.join(file_path, filename + file_ext)):
                                # Try appending a (#) suffix till a unique filename is found
                                filename = f'{options["post"]["content_filename"]}({counter})'
                                counter += 1
                        else:
                            logger.debug(
                                'Cannot ensure content_filename is unique '
                                'when adding to a remote transmission daemon.'
                            )

                        for file_id in main_list:
                            file_ext = os.path.splitext(file_list[file_id].name)[1]
                            logger.debug(
                                'File {} renamed to {}',
                                file_list[file_id].name,
                                filename + file_ext,
                            )
                            # change to below when set_files will allow setting name, more efficient to have one call
                            # fl[index]['name'] = os.path.basename(pathscrub(filename + file_ext).encode('utf-8'))
                            try:
                                self.client.rename_torrent_path(
                                    torrent_info.id,
                                    file_list[file_id].name,
                                    os.path.basename(str(pathscrub(filename + file_ext))),
                                )
                            except TransmissionError:
                                logger.error(
                                    'content_filename only supported with transmission 2.8+'
                                )

                    if options['post'].get('main_file_only') and main_id is not None:
                        # Set Unwanted Files
                        options['change']['files_unwanted'] = [
                            x for x in range(len(file_list)) if x not in dl_list
                        ]
                        options['change']['files_wanted'] = dl_list
                        logger.debug(
                            'Downloading {} of {} files in torrent.',
                            len(options['change']['files_wanted']),
                            len(file_list),
                        )
                    elif (
                        not options['post'].get('main_file_only') or main_id is None
                    ) and skip_files:
                        # If no main file and we want to skip files

                        if len(skip_list) >= len(file_list):
                            logger.debug(
                                'skip_files filter would cause no files to be downloaded; '
                                'including all files in torrent.'
                            )
                        else:
                            options['change']['files_unwanted'] = skip_list
                            options['change']['files_wanted'] = [
                                x for x in range(len(file_list)) if x not in skip_list
                            ]
                            logger.debug(
                                'Downloading {} of {} files in torrent.',
                                len(options['change']['files_wanted']),
                                len(file_list),
                            )

                # Set any changed file properties
                if list(options['change'].keys()):
                    self.client.change_torrent(torrent_info.id, 30, **options['change'])

                start_torrent = partial(self.client.start_torrent, [torrent_info.id])

                if config['action'] == 'add':
                    # if add_paused was defined and set to False start the torrent;
                    # prevents downloading data before we set what files we want
                    start_paused = (
                        options['post']['paused']
                        if 'paused' in options['post']
                        else not self.client.get_session().start_added_torrents
                    )
                    if start_paused:
                        self.client.stop_torrent(torrent_info.id)
                    else:
                        self.client.start_torrent(torrent_info.id)
                elif config['action'] in ('remove', 'purge'):
                    self.client.remove_torrent(
                        [torrent_info.id], delete_data=config['action'] == 'purge'
                    )
                    logger.info('{}d {} from transmission', config['action'], torrent_info.name)
                elif config['action'] == 'pause':
                    self.client.stop_torrent([torrent_info.id])
                    logger.info('paused {} in transmission', torrent_info.name)
                elif config['action'] == 'resume':
                    start_torrent()
                    logger.info('resumed {} in transmission', torrent_info.name)
                elif config['action'] == 'bypass_queue':
                    start_torrent(bypass_queue=True)
                    logger.info('resumed (bypass queue) {} in transmission', torrent_info.name)

            except TransmissionError as e:
                logger.opt(exception=True).debug('TransmissionError')
                logger.debug('Failed options dict: {}', options)
                msg = 'Error trying to {} {}, TransmissionError: {}'.format(
                    config['action'], entry['title'], e.message or 'N/A'
                )
                logger.error(msg)
                continue

    def _make_torrent_options_dict(self, config, entry):

        opt_dic = {}

        for opt_key in (
            'path',
            'add_paused',
            'honor_limits',
            'bandwidth_priority',
            'max_connections',
            'max_up_speed',
            'max_down_speed',
            'ratio',
            'main_file_only',
            'main_file_ratio',
            'magnetization_timeout',
            'include_subs',
            'content_filename',
            'include_files',
            'skip_files',
            'rename_like_files',
            'queue_position',
        ):
            # Values do not merge config with task
            # Task takes priority then config is used
            if opt_key in entry:
                opt_dic[opt_key] = entry[opt_key]
            elif opt_key in config:
                opt_dic[opt_key] = config[opt_key]

        options = {'add': {}, 'change': {}, 'post': {}}

        add = options['add']
        if opt_dic.get('path'):
            try:
                path = os.path.expanduser(entry.render(opt_dic['path']))
            except RenderError as e:
                logger.error('Error setting path for {}: {}', entry['title'], e)
            else:
                # Transmission doesn't like it when paths end in a separator
                path = path.rstrip('\\/')
                add['download_dir'] = pathscrub(path)
        # make sure we add it paused, will modify status after adding
        add['paused'] = True

        change = options['change']
        if 'bandwidth_priority' in opt_dic:
            change['bandwidthPriority'] = opt_dic['bandwidth_priority']
        if 'honor_limits' in opt_dic and not opt_dic['honor_limits']:
            change['honorsSessionLimits'] = False
        if 'max_up_speed' in opt_dic:
            change['uploadLimit'] = opt_dic['max_up_speed']
            change['uploadLimited'] = True
        if 'max_down_speed' in opt_dic:
            change['downloadLimit'] = opt_dic['max_down_speed']
            change['downloadLimited'] = True
        if 'max_connections' in opt_dic:
            change['peer_limit'] = opt_dic['max_connections']

        if 'ratio' in opt_dic:
            change['seedRatioLimit'] = opt_dic['ratio']
            if opt_dic['ratio'] == -1:
                # seedRatioMode:
                # 0 follow the global settings
                # 1 override the global settings, seeding until a certain ratio
                # 2 override the global settings, seeding regardless of ratio
                change['seedRatioMode'] = 2
            else:
                change['seedRatioMode'] = 1

        if 'queue_position' in opt_dic:
            change['queuePosition'] = opt_dic['queue_position']

        post = options['post']
        # set to modify paused status after
        if 'add_paused' in opt_dic:
            post['paused'] = opt_dic['add_paused']
        if 'main_file_only' in opt_dic:
            post['main_file_only'] = opt_dic['main_file_only']
        if 'main_file_ratio' in opt_dic:
            post['main_file_ratio'] = opt_dic['main_file_ratio']
        if 'magnetization_timeout' in opt_dic:
            post['magnetization_timeout'] = opt_dic['magnetization_timeout']
        if 'include_subs' in opt_dic:
            post['include_subs'] = opt_dic['include_subs']
        if 'content_filename' in opt_dic:
            try:
                post['content_filename'] = entry.render(opt_dic['content_filename'])
            except RenderError as e:
                logger.error('Unable to render content_filename {}: {}', entry['title'], e)
        if 'skip_files' in opt_dic:
            post['skip_files'] = opt_dic['skip_files']
            if not isinstance(post['skip_files'], list):
                post['skip_files'] = [post['skip_files']]
        if 'include_files' in opt_dic:
            post['include_files'] = opt_dic['include_files']
            if not isinstance(post['include_files'], list):
                post['include_files'] = [post['include_files']]
        if 'rename_like_files' in opt_dic:
            post['rename_like_files'] = opt_dic['rename_like_files']
        return options

    def on_task_learn(self, task, config):
        """ Make sure all temp files are cleaned up when entries are learned """
        # If download plugin is enabled, it will handle cleanup.
        if 'download' not in task.config:
            download = plugin.get('download', self)
            download.cleanup_temp_files(task)

    on_task_abort = on_task_learn


class PluginTransmissionClean(TransmissionBase):
    """
    DEPRECATED: A separate task using from_transmission and transmission with remove action should be used instead.

    Remove completed torrents from Transmission.

    Examples::

      clean_transmission: yes  # ignore both time and ratio

      clean_transmission:      # uses transmission's internal limits for idle time and seed ratio ( if defined )
        transmission_seed_limits: yes

      clean_transmission:      # matches time only
        finished_for: 2 hours

      clean_transmission:      # matches ratio only
        min_ratio: 0.5

      clean_transmission:      # matches time OR ratio
        finished_for: 2 hours
        min_ratio: 0.5

    Default values for the config elements::

      clean_transmission:
        host: localhost
        port: 9091
        enabled: yes
    """

    schema = {
        "deprecated": "The clean_transmission plugin is deprecated. Configure a new task using the from_transmission "
        "plugin as well as the transmission plugin using the remove or purge action.",
        "anyOf": [
            {"type": "boolean"},
            {
                "type": "object",
                "properties": {
                    "host": {"type": "string"},
                    "port": {"type": "integer"},
                    "netrc": {"type": "string", "format": "file"},
                    "username": {"type": "string"},
                    "password": {"type": "string"},
                    "enabled": {"type": "boolean"},
                    "min_ratio": {"type": "number"},
                    "finished_for": {"type": "string", "format": "interval"},
                    "transmission_seed_limits": {"type": "boolean"},
                    "delete_files": {"type": "boolean"},
                    "tracker": {"type": "string", "format": "regex"},
                    "preserve_tracker": {"type": "string", "format": "regex"},
                    "directories": {
                        "type": "array",
                        "items": {"type": "string", "format": "regex"},
                    },
                },
                "additionalProperties": False,
            },
        ],
    }

    def on_task_exit(self, task, config):
        config = self.prepare_config(config)
        if not config['enabled'] or task.options.learn:
            return
        if not self.client:
            self.client = self.create_rpc_client(config)
        tracker_re = re.compile(config['tracker'], re.IGNORECASE) if 'tracker' in config else None
        preserve_tracker_re = (
            re.compile(config['preserve_tracker'], re.IGNORECASE)
            if 'preserve_tracker' in config
            else None
        )

        session = self.client.get_session()

        remove_ids = []
        for torrent in self.client.get_torrents():
            logger.verbose(
                'Torrent "{}": status: "{}" - ratio: {} - date added: {}',
                torrent.name,
                torrent.status,
                torrent.ratio,
                torrent.date_added,
            )
            downloaded, dummy = self.torrent_info(torrent, config)
            if not downloaded:
                continue
            if config.get('transmission_seed_limits'):
                seed_ratio_ok, idle_limit_ok = self.check_seed_limits(torrent, session)
                if not seed_ratio_ok or not idle_limit_ok:
                    continue
            if 'min_ratio' in config:
                if torrent.ratio < config['min_ratio']:
                    continue
            if 'finished_for' in config:
                # done date might be invalid if this torrent was added to transmission when already completed
                started_seeding = datetime.fromtimestamp(max(torrent.addedDate, torrent.doneDate))
                if started_seeding + parse_timedelta(config['finished_for']) > datetime.now():
                    continue
            tracker_hosts = (
                urlparse(tracker['announce']).hostname for tracker in torrent.trackers
            )
            if 'tracker' in config:
                if not any(tracker_re.search(tracker) for tracker in tracker_hosts):
                    continue
            if 'preserve_tracker' in config:
                if any(preserve_tracker_re.search(tracker) for tracker in tracker_hosts):
                    continue
            if config.get('directories'):
                if not any(
                    re.search(d, torrent.downloadDir, re.IGNORECASE) for d in config['directories']
                ):
                    continue
            if task.options.test:
                logger.info('Would remove finished torrent `{}` from transmission', torrent.name)
                continue
            logger.info('Removing finished torrent `{}` from transmission', torrent.name)
            remove_ids.append(torrent.id)
        if remove_ids:
            self.client.remove_torrent(remove_ids, config.get('delete_files'))


@event('plugin.register')
def register_plugin():
    plugin.register(PluginTransmission, 'transmission', api_ver=2)
    plugin.register(PluginTransmissionInput, 'from_transmission', api_ver=2)
    plugin.register(PluginTransmissionClean, 'clean_transmission', api_ver=2)
